# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for input field widgets."""

from typing import ClassVar, override

import lxml
from django.template import Context, engines
from django.test import RequestFactory

import debusine.web.views.input_fields as input_widgets
from debusine.artifacts.models import CollectionCategory
from debusine.db.models import Artifact, Collection
from debusine.db.tasks import inputs as dbinputs
from debusine.db.tasks.models import (
    DBArtifactInfo,
    DBBaseTaskInputArtifact,
    DBInputArtifactMultiple,
    DBInputCollectionSingle,
)
from debusine.tasks import inputs
from debusine.tasks.models import (
    BaseTaskInputArtifact,
    ExtraExternalRepository,
    InputArtifactSingle,
    InputCollectionSingle,
    LookupMultiple,
)
from debusine.test.django import TestCase


class WidgetTestCase(TestCase):
    """Base test case for testing widgets."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.request_factory = RequestFactory()
        self.request = self.request_factory.get("/")
        self.context = Context({"request": self.request})
        self.template = engines["django"].from_string("")
        self.context.template = self.template.template  # type: ignore[attr-defined]


class RenderArtifactTests(WidgetTestCase):
    """Test :py:func:`render_artifact`."""

    artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.artifact = cls.playground.create_minimal_source_package_artifact()

    def assertRendersActive(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> None:
        root = self.assertHasElement(tree, "body/div")
        a = root.div[0].a
        self.assertEqual(a.get("href"), self.artifact.get_absolute_url())
        shortcuts = root.div[1]
        self.assertEqual(len(shortcuts.a), 2)

    def assertRendersExpired(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> None:
        root = self.assertHasElement(tree, "body/div")
        self.assertHasElement(root.div[0], "span[@class='expired']")
        # There are no shortcuts
        self.assertEqual(len(root.div), 1)

    def test_render_artifactinfo(self) -> None:
        item = InputArtifactSingle(
            artifact_id=self.artifact.id,
            category=self.artifact.category,
            data=self.artifact.create_data(),
            label="label",
            lookup="lookup",
        )
        tree = self.assertHTMLValid(
            input_widgets.render_artifact(item, self.context)
        )
        self.assertRendersActive(tree)

    def test_render_artifactinfo_expired(self) -> None:
        item = InputArtifactSingle(
            artifact_id=-1,
            category=self.artifact.category,
            data=self.artifact.create_data(),
            label="label",
            lookup="lookup",
        )
        tree = self.assertHTMLValid(
            input_widgets.render_artifact(item, self.context)
        )
        self.assertRendersExpired(tree)

    def test_render_basetaskinputartifact(self) -> None:
        item = BaseTaskInputArtifact(
            artifact_id=self.artifact.id,
            category=self.artifact.category,
            data=self.artifact.create_data(),
        )
        tree = self.assertHTMLValid(
            input_widgets.render_artifact(item, self.context)
        )
        self.assertRendersActive(tree)

    def test_render_basetaskinputartifact_expired(self) -> None:
        item = BaseTaskInputArtifact(
            artifact_id=-1,
            category=self.artifact.category,
            data=self.artifact.create_data(),
        )
        tree = self.assertHTMLValid(
            input_widgets.render_artifact(item, self.context)
        )
        self.assertRendersExpired(tree)

    def test_render_dbartifactinfo(self) -> None:
        item = DBArtifactInfo.from_artifact(
            self.artifact, label="label", lookup="lookup"
        )
        tree = self.assertHTMLValid(
            input_widgets.render_artifact(item, self.context)
        )
        self.assertRendersActive(tree)

    def test_render_dbartifactinfo_unchached(self) -> None:
        item = DBArtifactInfo.from_artifact(
            self.artifact, label="label", lookup="lookup"
        )
        item._artifact = None
        tree = self.assertHTMLValid(
            input_widgets.render_artifact(item, self.context)
        )
        self.assertRendersActive(tree)

    def test_render_dbartifactinfo_expired(self) -> None:
        item = DBArtifactInfo(
            artifact_id=-1,
            category=self.artifact.category,
            data=self.artifact.create_data(),
            label="label",
            lookup="lookup",
        )
        tree = self.assertHTMLValid(
            input_widgets.render_artifact(item, self.context)
        )
        self.assertRendersExpired(tree)

    def test_render_dbbasetaskinputartifact(self) -> None:
        item = DBBaseTaskInputArtifact.from_artifact(self.artifact)
        tree = self.assertHTMLValid(
            input_widgets.render_artifact(item, self.context)
        )
        self.assertRendersActive(tree)

    def test_render_dbbasetaskinputartifact_uncached(self) -> None:
        item = DBBaseTaskInputArtifact.from_artifact(self.artifact)
        item._artifact = None
        tree = self.assertHTMLValid(
            input_widgets.render_artifact(item, self.context)
        )
        self.assertRendersActive(tree)

    def test_render_dbbasetaskinputartifact_expired(self) -> None:
        item = BaseTaskInputArtifact(
            artifact_id=-1,
            category=self.artifact.category,
            data=self.artifact.create_data(),
        )
        tree = self.assertHTMLValid(
            input_widgets.render_artifact(item, self.context)
        )
        self.assertRendersExpired(tree)


class RenderCollectionTests(WidgetTestCase):
    """Test :py:func:`render_collection`."""

    collection: ClassVar[Collection]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.collection = cls.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )

    def assertRendersActive(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> None:
        a = self.assertHasElement(tree, "body/a")
        self.assertEqual(a.get("href"), self.collection.get_absolute_url())

    def assertRendersExpired(
        self, tree: lxml.objectify.ObjectifiedElement
    ) -> None:
        root = self.assertHasElement(tree, "body")
        self.assertTextContentEqual(
            root,
            f"#-1 {self.collection.name}@{self.collection.category}",
        )

    def test_render_inputcollectionsingle(self) -> None:
        item = InputCollectionSingle(
            collection_id=self.collection.id,
            name=self.collection.name,
            category=self.collection.category,
            scope_name=self.collection.workspace.scope.name,
            workspace_name=self.collection.workspace.name,
            data=self.collection.data,
            label="label",
            lookup="lookup",
        )
        tree = self.assertHTMLValid(
            input_widgets.render_collection(item, self.context)
        )
        self.assertRendersActive(tree)

    def test_render_inputcollectionsingle_uncached(self) -> None:
        item = InputCollectionSingle(
            collection_id=self.collection.id,
            name=self.collection.name,
            category=self.collection.category,
            scope_name=self.collection.workspace.scope.name,
            workspace_name=self.collection.workspace.name,
            data=self.collection.data,
            label="label",
            lookup="lookup",
        )
        tree = self.assertHTMLValid(
            input_widgets.render_collection(item, self.context)
        )
        self.assertRendersActive(tree)

    def test_render_inputcollectionsingle_expired(self) -> None:
        item = InputCollectionSingle(
            collection_id=-1,
            name=self.collection.name,
            category=self.collection.category,
            scope_name=self.collection.workspace.scope.name,
            workspace_name=self.collection.workspace.name,
            data=self.collection.data,
            label="label",
            lookup="lookup",
        )
        tree = self.assertHTMLValid(
            input_widgets.render_collection(item, self.context)
        )
        self.assertRendersExpired(tree)

    def test_render_dbinputcollectionsingle(self) -> None:
        item = DBInputCollectionSingle.from_collection(
            self.collection, label="label", lookup="lookup"
        )
        tree = self.assertHTMLValid(
            input_widgets.render_collection(item, self.context)
        )
        self.assertRendersActive(tree)

    def test_render_dbinputcollectionsingle_uncached(self) -> None:
        item = DBInputCollectionSingle(
            collection_id=self.collection.id,
            name=self.collection.name,
            category=self.collection.category,
            scope_name=self.collection.workspace.scope.name,
            workspace_name=self.collection.workspace.name,
            data=self.collection.data,
            label="label",
            lookup="lookup",
        )
        item._collection = None
        tree = self.assertHTMLValid(
            input_widgets.render_collection(item, self.context)
        )
        self.assertRendersActive(tree)

    def test_render_dbinputcollectionsingle_expired(self) -> None:
        item = DBInputCollectionSingle(
            collection_id=-1,
            name=self.collection.name,
            category=self.collection.category,
            scope_name=self.collection.workspace.scope.name,
            workspace_name=self.collection.workspace.name,
            data=self.collection.data,
            label="label",
            lookup="lookup",
        )
        tree = self.assertHTMLValid(
            input_widgets.render_collection(item, self.context)
        )
        self.assertRendersExpired(tree)


class RenderLookupTests(WidgetTestCase):
    """Test :py:func:`render_lookup`."""

    def test_int(self) -> None:
        rendered = input_widgets.render_lookup(42)
        self.assertEqual(rendered, "<code>42</code>")

    def test_str(self) -> None:
        rendered = input_widgets.render_lookup("foo")
        self.assertEqual(rendered, "<code>foo</code>")

    def test_multiple(self) -> None:
        lookup = LookupMultiple((1, 2, 3))
        tree = self.assertHTMLValid(input_widgets.render_lookup(lookup))
        self.assertYAMLContentEqual(tree, [1, 2, 3])

    def test_other(self) -> None:
        tree = self.assertHTMLValid(input_widgets.render_lookup({"foo": "bar"}))  # type: ignore[arg-type]
        self.assertYAMLContentEqual(tree, {"foo": "bar"})


class InputFieldWidgetTests(WidgetTestCase):
    """Test :py:class:`InputFieldWidget`."""

    def test_render_details_no_help_text(self) -> None:
        field = inputs.SingleInput(field="test")
        field.name = "test"
        widget = input_widgets.InputFieldWidget(field, None)
        tree = self.assertHTMLValid(widget.render_details())
        root = self.assertHasElement(tree, "body")
        self.assertFalse(root.xpath("p"))
        dl = self.assertHasElement(root, "dl")
        self.assertTextContentEqual(dl.dt[0], "Defined as:")
        self.assertTextContentEqual(dl.dd[0], "test")
        self.assertEqual(len(dl.dt), 1)

    def test_render_details_append_fullstop(self) -> None:
        field = inputs.SingleInput(field="test", help_text="Help text")
        field.name = "test"
        widget = input_widgets.InputFieldWidget(field, None)
        tree = self.assertHTMLValid(widget.render_details())
        root = self.assertHasElement(tree, "body")
        p = self.assertHasElement(root, "p")
        self.assertTextContentEqual(p, "Help text.")

    def test_render_details_append_fullstop_only_if_missing(self) -> None:
        field = inputs.SingleInput(field="test", help_text="Help text.")
        field.name = "test"
        widget = input_widgets.InputFieldWidget(field, None)
        tree = self.assertHTMLValid(widget.render_details())
        root = self.assertHasElement(tree, "body")
        p = self.assertHasElement(root, "p")
        self.assertTextContentEqual(p, "Help text.")

    def test_render_value_rows(self) -> None:
        artifact = self.playground.create_minimal_source_package_artifact()
        value = DBArtifactInfo.from_artifact(
            artifact, lookup="lookup", label="test"
        )
        field = inputs.SingleInput(field="test")
        field.name = "test"
        widget = input_widgets.InputFieldWidget(field, value)
        rows = list(widget.render_value_rows(self.context))
        self.assertEqual(len(rows), 1)
        tree = self.assertHTMLValid(rows[0])
        self.assertYAMLContentEqual(tree, value.model_dump())


class ArtifactInputFieldWidgetTests(WidgetTestCase):
    """Test :py:class:`ArtifactInputFieldWidget`."""

    artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.artifact = cls.playground.create_minimal_source_package_artifact()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.value = DBArtifactInfo.from_artifact(
            self.artifact, label="test", lookup=self.artifact.id
        )

        class Container(inputs.TaskFieldContainer):
            test = inputs.OptionalSingleInput()

        self.container = Container()
        self.container.test = self.value

    def test_create(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertIsInstance(widget, input_widgets.ArtifactInputFieldWidget)

    def test_render(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        tree = self.assertHTMLValid(widget.render(self.context))
        root = self.assertHasElement(tree, "body")
        self.assertEqual(len(root.tr), 1)
        self.assertIsNone(root.tr[0].td[0].get("rowspan"))

    def test_render_none(self) -> None:
        self.container.test = None
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertEqual(widget.render(self.context), "")

    def test_render_details_rows(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        rows = list(widget.render_details_rows())
        self.assertEqual(
            rows,
            [
                ("Defined as", "<code>test</code>"),
                ("Task input field", "<code>test</code>"),
                ("Lookup", input_widgets.render_lookup(self.artifact.id)),
            ],
        )

    def test_render_details_rows_empty(self) -> None:
        self.container.test = None
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        rows = list(widget.render_details_rows())
        self.assertEqual(rows, [("Defined as", "<code>test</code>")])


class InputArtifactMultipleFieldWidgetTests(WidgetTestCase):
    """Test :py:class:`InputArtifactMultipleFieldWidget`."""

    artifacts: ClassVar[list[Artifact]]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.artifacts = [
            cls.playground.create_minimal_source_package_artifact(
                name="hello1"
            ),
            cls.playground.create_minimal_source_package_artifact(
                name="hello2"
            ),
            cls.playground.create_minimal_source_package_artifact(
                name="hello3"
            ),
        ]

    @override
    def setUp(self) -> None:
        super().setUp()
        self.value = DBInputArtifactMultiple(
            label="test",
            lookup=[a.id for a in self.artifacts],
            artifacts=[
                DBBaseTaskInputArtifact.from_artifact(a) for a in self.artifacts
            ],
        )

        class Container(inputs.TaskFieldContainer):
            test = inputs.MultiInput()

        self.container = Container()
        self.container.test = self.value

    def test_create(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertIsInstance(
            widget, input_widgets.InputArtifactMultipleFieldWidget
        )

    def test_render(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        tree = self.assertHTMLValid(widget.render(self.context))
        root = self.assertHasElement(tree, "body")
        self.assertEqual(len(root.tr), 3)
        self.assertEqual(root.tr[0].td[0].get("rowspan"), "3")

    def test_render_empty(self) -> None:
        self.container.test = DBInputArtifactMultiple(
            label="test", lookup=[a.id for a in self.artifacts], artifacts=[]
        )

        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertEqual(widget.render(self.context), "")


class ArtifactInputListFieldWidgetTests(WidgetTestCase):
    """Test :py:class:`ArtifactInputListFieldWidget`."""

    artifacts: ClassVar[list[Artifact]]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.artifacts = [
            cls.playground.create_minimal_source_package_artifact(
                name="hello1"
            ),
            cls.playground.create_minimal_source_package_artifact(
                name="hello2"
            ),
            cls.playground.create_minimal_source_package_artifact(
                name="hello3"
            ),
        ]

    @override
    def setUp(self) -> None:
        super().setUp()
        self.value: list[InputArtifactSingle] = [
            DBArtifactInfo.from_artifact(a, lookup=a.id, label=f"test[{idx}]")
            for idx, a in enumerate(self.artifacts)
        ]

        class Container(inputs.TaskFieldContainer):
            test = inputs.SingleInputList()

        self.container = Container()
        self.container.test = self.value

    def test_create(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertIsInstance(
            widget, input_widgets.ArtifactInputListFieldWidget
        )

    def test_render(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        tree = self.assertHTMLValid(widget.render(self.context))
        root = self.assertHasElement(tree, "body")
        self.assertEqual(len(root.tr), 3)
        self.assertEqual(root.tr[0].td[0].get("rowspan"), "3")

    def test_render_empty(self) -> None:
        self.container.test = []
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertEqual(widget.render(self.context), "")

    def test_render_details_rows(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        rows = list(widget.render_details_rows())
        self.assertEqual(
            rows,
            [
                ("Defined as", "<code>test</code>"),
                (
                    "Task input field <code>test[0]</code>",
                    input_widgets.render_lookup(self.artifacts[0].id),
                ),
                (
                    "Task input field <code>test[1]</code>",
                    input_widgets.render_lookup(self.artifacts[1].id),
                ),
                (
                    "Task input field <code>test[2]</code>",
                    input_widgets.render_lookup(self.artifacts[2].id),
                ),
            ],
        )

    def test_render_details_rows_empty(self) -> None:
        self.container.test = []
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        rows = list(widget.render_details_rows())
        self.assertEqual(rows, [("Defined as", "<code>test</code>")])


class CollectionInputFieldWidgetTests(WidgetTestCase):
    """Test :py:class:`CollectionInputFieldWidget`."""

    collection: ClassVar[Collection]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.collection = cls.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.value = DBInputCollectionSingle.from_collection(
            self.collection, label="test", lookup=self.collection.id
        )

        class Container(inputs.TaskFieldContainer):
            test = dbinputs.CollectionInput()

        self.container = Container()
        self.container.test = self.value

    def test_create(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertIsInstance(widget, input_widgets.CollectionInputFieldWidget)

    def test_render(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        tree = self.assertHTMLValid(widget.render(self.context))
        root = self.assertHasElement(tree, "body")
        self.assertEqual(len(root.tr), 1)
        self.assertIsNone(root.tr[0].td[0].get("rowspan"))

    def test_render_none(self) -> None:
        self.container.test = None  # type: ignore[assignment]
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertEqual(widget.render(self.context), "")

    def test_render_details_rows(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        rows = list(widget.render_details_rows())
        self.assertEqual(
            rows,
            [
                ("Defined as", "<code>test</code>"),
                ("Task input field", "<code>test</code>"),
                ("Lookup", input_widgets.render_lookup(self.collection.id)),
            ],
        )

    def test_render_details_rows_empty(self) -> None:
        self.container.test = None  # type: ignore[assignment]
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        rows = list(widget.render_details_rows())
        self.assertEqual(rows, [("Defined as", "<code>test</code>")])


class ExtraExternalRepositoryFieldWidgetTests(WidgetTestCase):
    """Test :py:class:`ExtraExternalRepositoryFieldWidget`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.value: list[ExtraExternalRepository] = [
            ExtraExternalRepository(
                url="http://deb.debian.org/debian",
                suite="sid",
                components=["main", "non-free-firmware"],
            ),
            ExtraExternalRepository(
                url="http://deb.debian.org/debian",
                suite="experimental",
                components=["main"],
            ),
        ]

        class Container(inputs.TaskFieldContainer):
            test = inputs.ExtraRepositoriesInput()

        self.container = Container()
        self.container.test = self.value

    def test_create(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertIsInstance(
            widget, input_widgets.ExtraExternalRepositoryFieldWidget
        )

    def test_render(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        tree = self.assertHTMLValid(widget.render(self.context))
        root = self.assertHasElement(tree, "body")
        self.assertEqual(len(root.tr), 2)
        self.assertEqual(root.tr[0].td[0].get("rowspan"), "2")

    def test_render_none(self) -> None:
        self.container.test = None
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertEqual(widget.render(self.context), "")


class FallbackWidgetTests(WidgetTestCase):
    """Test instantiating a widget for an unsupported field."""

    suite: ClassVar[Collection]
    archive: ClassVar[Collection]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.suite = cls.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        cls.archive = cls.playground.create_singleton_collection(
            category=CollectionCategory.ARCHIVE
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.value: tuple[
            InputCollectionSingle, InputCollectionSingle | None
        ] = (
            DBInputCollectionSingle.from_collection(
                self.suite, lookup=self.suite.id, label="test"
            ),
            DBInputCollectionSingle.from_collection(
                self.archive, lookup="_@debian:archive", label="test"
            ),
        )

        class Container(inputs.TaskFieldContainer):
            test = inputs.SuiteArchiveInput()

        self.container = Container()
        self.container.test = self.value

    def test_create(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        self.assertIsInstance(widget, input_widgets.InputFieldWidget)

    def test_render(self) -> None:
        widget = input_widgets.InputFieldWidget.create(
            self.container, self.container.inputs["test"]
        )
        tree = self.assertHTMLValid(widget.render(self.context))
        root = self.assertHasElement(tree, "body")
        self.assertEqual(len(root.tr), 1)
        self.assertIsNone(root.tr[0].td[0].get("rowspan"))
        assert self.value[1] is not None
        self.assertYAMLContentEqual(
            root.tr.td[1],
            [self.value[0].model_dump(), self.value[1].model_dump()],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Image handling for executor backends."""

import datetime as dt
import logging
from pathlib import Path
from shutil import rmtree

from fasteners import InterProcessLock

from debusine.client.debusine import Debusine
from debusine.client.models import ArtifactResponse
from debusine.tasks.executors.base import ExecutorImageCategory, _backends
from debusine.tasks.models import (
    BackendType,
    ImageCacheUsageLog,
    ImageCacheUsageLogEntry,
)

CACHE_SIZE = 10
USAGE_LOG = "usage.json"

log = logging.getLogger(__name__)
# Alias datetime.now to make it easier to mock
now = dt.datetime.now


class ImageCache:
    """
    Download image artifacts to a local cache.

    Processed versions of an image may be stored for a backend.
    """

    image_cache_path: Path = Path.home() / "system-images"
    image_category: ExecutorImageCategory
    backend: BackendType | None = None
    workspace: str

    def __init__(
        self,
        debusine_api: Debusine,
        image_category: ExecutorImageCategory,
        workspace: str,
    ) -> None:
        """
        Instantiate an ImageCache.

        :param debusine_api: object to use the debusine client API
        :param image_category: artifact category of supported images
        """
        self.debusine_api = debusine_api
        self.image_category = image_category
        self.workspace = workspace

    def cache_artifact_path(self, artifact_id: int) -> Path:
        """Return the path to the image cache pool for artifact_id."""
        return self.image_cache_path / str(artifact_id)

    def cache_artifact_image_path(
        self,
        artifact_id: int,
        filename: str,
        backend: BackendType | None = None,
    ) -> Path:
        """Return the path to the cached image for filename, for backend."""
        cache_path = self.cache_artifact_path(artifact_id)
        if backend:
            cache_path = cache_path / "backends" / backend
        return cache_path / filename

    def image_artifact(self, artifact_id: int) -> ArtifactResponse:
        """Retrieve an Artifact."""
        r = self.debusine_api.artifact_get(
            artifact_id, workspace=self.workspace
        )
        if r.category != self.image_category.value:
            raise ValueError(
                f"Unexpected artifact type {r.category}; expected "
                f"{self.image_category}"
            )
        return r

    def _image_lock(self, artifact_id: int) -> InterProcessLock:
        """Context Manager to hold a RW lock for working on a cached image."""
        cache_path = self.cache_artifact_path(artifact_id)
        cache_path.mkdir(exist_ok=True)
        return InterProcessLock(cache_path / ".lock")

    def process_downloaded_image(self, artifact: ArtifactResponse) -> None:
        """
        Handle any necessary post-download processing of the image.

        Locate the downloaded image using cache_artifact_image_path() and store
        it for a backend, if the result needs to be stored in the cache.

        This only needs to be done once, after the image is first downloaded.
        """

    def image_in_cache(self, artifact: ArtifactResponse) -> bool:
        """Return whether or not artifact is available in the local cache."""
        return self.cache_artifact_image_path(
            artifact.id, artifact.data["filename"]
        ).exists()

    def _download_image_artifact(
        self, artifact_id: int, filename: str, destination: Path
    ) -> None:
        """Atomically download filename from artifact to destination."""
        temp_path = destination.parent / (".download.tmp." + destination.name)
        temp_path.unlink(missing_ok=True)
        self.debusine_api.download_artifact_file(
            artifact_id,
            path_in_artifact=filename,
            destination=temp_path,
            workspace=self.workspace,
        )
        temp_path.rename(destination)

    def download_image(self, artifact: ArtifactResponse) -> Path:
        """
        Make the image available locally.

        Fetch the image from artifact storage, if it isn't already available,
        and make it available locally.

        Return a path to the image or name, as appropriate for the backend.
        """
        self.image_cache_path.mkdir(exist_ok=True)

        filename = artifact.data["filename"]

        # Protect our image-cache directory structure
        if filename.startswith(".") or filename in ("backends", USAGE_LOG):
            raise ValueError("Image has an invalid filename")

        path = self.cache_artifact_image_path(artifact.id, filename)

        with self._image_lock(artifact.id):
            if not self.image_in_cache(artifact):
                if not path.exists():
                    self._download_image_artifact(artifact.id, filename, path)
                self.process_downloaded_image(artifact)
            self.record_image_use(artifact)
        self.clean_up_old_images()
        return path

    def record_image_use(self, artifact: ArtifactResponse) -> None:
        """
        Record the use of an image.

        Executed while the image lock is held.
        """
        record_file = self.cache_artifact_image_path(artifact.id, USAGE_LOG)
        if record_file.exists():
            with record_file.open("r") as f:
                record = ImageCacheUsageLog.model_validate_json(f.read())
        else:
            record = ImageCacheUsageLog()

        if self.backend:
            record.backends.add(self.backend)

        record.usage = record.usage[-CACHE_SIZE + 1 :]
        record.usage.append(
            ImageCacheUsageLogEntry(
                filename=artifact.data["filename"],
                backend=self.backend,
                timestamp=now(dt.UTC),
            )
        )

        tmp_file = self.cache_artifact_image_path(
            artifact.id, f".tmp.{USAGE_LOG}"
        )
        with tmp_file.open("w") as f:
            f.write(record.model_dump_json())
        tmp_file.replace(record_file)

    def clean_up_old_images(self) -> None:
        """
        Purge old images from the cache.

        This implements simple LRU eviction, evicting all backend versions at
        once.
        """
        for artifact_id, backends in self._find_least_used_images():
            with self._image_lock(artifact_id):
                for backend in backends:
                    try:
                        backend_cls = _backends[backend]
                        if hasattr(backend_cls, "clean_up_image"):
                            backend_cls.clean_up_image(artifact_id)
                    except Exception:
                        # Clean-up is in a critical code path, don't fail, just
                        # log errors. If the log gets corrupted, we'll attempt
                        # a full clean-up including backends that may not exist.
                        log.warning(
                            "Failed to clean up %i backend %s",
                            artifact_id,
                            backend,
                            exc_info=True,
                        )
                rmtree(self.cache_artifact_path(artifact_id))

    def _find_least_used_images(self) -> list[tuple[int, set[BackendType]]]:
        """
        Find the least recently used images in the cache.

        Return a list of (artifact_id, [backends]) tuples.
        """
        last_used: dict[int, dt.datetime] = {}
        backends: dict[int, set[BackendType]] = {}
        unknown: list[int] = []

        for path in self.image_cache_path.iterdir():
            if not path.name.isdigit():
                continue
            artifact_id = int(path.name)
            with self._image_lock(artifact_id):
                record_file = path / USAGE_LOG
                if not record_file.exists():
                    log.warning("No usage log for artifact %i", artifact_id)
                    unknown.append(artifact_id)
                    continue
                try:
                    with record_file.open("r") as f:
                        record = ImageCacheUsageLog.model_validate_json(
                            f.read()
                        )
                except Exception:
                    log.warning(
                        "Failed to parse log for artifact %i", artifact_id
                    )
                    unknown.append(artifact_id)
                    continue
                last_used[artifact_id] = record.usage[-1].timestamp
                backends[artifact_id] = record.backends

        most_recently_used = sorted(
            last_used.items(), key=lambda item: item[1], reverse=True
        )

        least_recently_used = [
            (artifact_id, backends[artifact_id])
            for artifact_id, _timestamp in most_recently_used[CACHE_SIZE:]
        ]
        # Assume the worst

        least_recently_used += [
            (artifact_id, set(_backends.keys())) for artifact_id in unknown
        ]

        return least_recently_used



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
debusine-signing register_pkcs11_static_key command.

This registers a static key stored on a PKCS#11 token.
"""

from pathlib import Path
from typing import Any, override

from django.core.exceptions import ValidationError
from django.core.management import CommandError, CommandParser

from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.signing.db.models import AuditLog, Key
from debusine.signing.models import ProtectedKeyPKCS11Static


class Command(DebusineBaseCommand):
    """Command to register a static key stored on a PKCS#11 token."""

    help = "Register a static key stored on a PKCS#11 token"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        """Add CLI arguments for the register_pkcs11_static_key command."""
        parser.add_argument(
            "purpose",
            choices=list(Key.Purpose),
            help="The purpose of the key.",
        )
        parser.add_argument("pkcs11_uri", help="The pkcs11: URI for the key.")
        parser.add_argument(
            "certificate_file",
            type=Path,
            help="File path to a certificate for the key.",
        )
        parser.add_argument("description", help="Description of the key.")

    @override
    def handle(self, *args: Any, **options: Any) -> None:
        """Register a key."""
        purpose = options["purpose"]
        pkcs11_uri = options["pkcs11_uri"]
        public_key = options["certificate_file"].read_bytes()
        description = options["description"]
        fingerprint = Key.objects.get_fingerprint(
            purpose=purpose, public_key=public_key
        )

        try:
            key = Key(
                purpose=purpose,
                fingerprint=fingerprint,
                private_key=ProtectedKeyPKCS11Static.create(
                    pkcs11_uri=pkcs11_uri
                ).model_dump(),
                public_key=public_key,
            )
            key.full_clean()
            key.save()
            AuditLog.objects.create(
                event=AuditLog.Event.REGISTER,
                purpose=purpose,
                fingerprint=fingerprint,
                data={"description": description},
            )
        except ValidationError as exc:
            raise CommandError(
                "Error creating key: " + "\n".join(exc.messages), returncode=3
            )

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the FileUploader class."""

import logging
import mmap
from collections.abc import Callable
from typing import override
from unittest import mock

import requests
import responses

from debusine.client import exceptions
from debusine.client.file_uploader import (
    FileUploadError,
    FileUploader,
    UnexpectedStatusCodeError,
)
from debusine.test import TestCase
from debusine.utils import parse_range_header


def body_matcher(
    expected_body: bytes | None,
) -> Callable[..., tuple[bool, str]]:  # pragma: no cover
    """Matcher for the body's content of a request. Used by responses.add."""

    def match(request: requests.models.PreparedRequest) -> tuple[bool, str]:
        """Return True if request_body matches expected_body."""
        if isinstance(request.body, mmap.mmap):
            position = request.body.tell()
            contents = request.body.read()
            request.body.seek(position)
        else:
            contents = request.body

        matched = expected_body == contents

        if matched:
            reason = "Bodies matched"
        else:
            reason = "Bodies are different"

        return matched, reason

    return match


class UnexpectedStatusCodeErrorTests(TestCase):
    """Tests for UnexpectedStatusCodeError."""

    @override
    def setUp(self) -> None:
        """Set up object with fake response."""
        super().setUp()
        mocked_request = mock.create_autospec(spec=requests.models.Request)
        mocked_request.method = "PUT"
        mocked_request.headers = {}
        mocked_request.url = "http://localhost:8088/test"

        self.mocked_response = mock.create_autospec(
            spec=requests.models.Response
        )
        self.mocked_response.request = mocked_request
        self.mocked_response.status_code = requests.codes.bad_request
        self.mocked_response.text = "Error from the server"

    def test_str_without_content_range(self) -> None:
        """Test for the human-readable string: content-range was not sent."""
        status_code_error = UnexpectedStatusCodeError(self.mocked_response)

        mocked_response = self.mocked_response
        mocked_request = self.mocked_response.request

        self.assertEqual(
            str(status_code_error),
            f"Client HTTP {mocked_request.method} {mocked_request.url} "
            f"Received unexpected status code: {mocked_response.status_code} "
            f"Response body:\n{mocked_response.text}",
        )

    def test_str_with_content_range(self) -> None:
        """Test for the human-readable string: content-range was in headers."""
        self.mocked_response.request.headers = {"Content-Range": "*/4"}

        status_code_error = UnexpectedStatusCodeError(self.mocked_response)

        mocked_response = self.mocked_response
        mocked_request = self.mocked_response.request

        self.assertEqual(
            str(status_code_error),
            f"Client HTTP {mocked_request.method} {mocked_request.url} "
            f"with Content-Range {mocked_request.headers['Content-Range']} "
            f"Received unexpected status code: {mocked_response.status_code} "
            f"Response body:\n{mocked_response.text}",
        )


class FileUploaderTests(TestCase):
    """Tests for the debusine command line interface."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.url = "https://localhost:8011/1.0/artifact/2/README.md"
        self.token = "21efb7587bd10b8eb2c56b4280b4c4a92a9"
        self.scope = "scope"

        self.file_uploader = FileUploader(
            self.token, scope=self.scope, logger=logging.getLogger(__name__)
        )

        self.file_content = b"test-content"
        self.file_path = self.create_temporary_file(contents=self.file_content)

    def add_partial_response(self, status: int, response_range: str) -> None:
        """
        Add response: for PUT in self.url with empty body.

        :param status: response's status code.
        :param response_range: response's "Range" header.
        """
        responses.add(
            responses.PUT,
            self.url,
            status=status,
            adding_headers={"Range": f"bytes={response_range}"},
            match=[body_matcher(None)],
        )

    def setup_partial_file_uploaded(
        self,
        /,
        uploaded_size: int,
        status_code_file_status: int = requests.codes.partial_content,
        status_code_upload: int = requests.codes.created,
    ) -> tuple[int, str]:
        """Set up responses for a file partially uploaded."""
        file_size = len(self.file_content)

        self.add_partial_response(status_code_file_status, f"0-{uploaded_size}")

        responses.add(
            responses.PUT,
            self.url,
            status=status_code_upload,
            match=[
                body_matcher(self.file_content[uploaded_size:file_size]),
            ],
        )

        # Missing part to upload
        content_range = f"bytes {uploaded_size}-{file_size - 1}/{file_size}"
        return file_size - uploaded_size, content_range

    @responses.activate
    def test_authentication(self) -> None:
        """Client sets the correct headers to authenticate the request."""
        responses.add(responses.PUT, self.url)
        self.file_uploader.upload(self.file_path, self.url)

        request = responses.calls[0].request

        self.assertEqual(request.headers["Token"], self.token)
        self.assertEqual(request.headers["X-Debusine-Scope"], self.scope)

    @responses.activate
    def test_file_uploader_check_file_was_completed(self) -> None:
        """
        Client sends PUT Content-Range: */file_size. Server return HTTP 200.

        The client does not upload anything.
        """  # noqa: RST213
        responses.add(responses.PUT, self.url)
        self.file_uploader.upload(self.file_path, self.url)

        file_size = len(self.file_content)

        request = responses.calls[0].request

        # The client requested to know the missing part of the file
        self.assertEqual(
            request.headers["Content-Range"], f"bytes */{file_size}"
        )

        # The response was HTTP 200 (requests.codes.ok) so no more HTTP
        # requests: nothing to do
        self.assertEqual(len(responses.calls), 1)

        # Nothing was uploaded to the server
        self.assertIsNone(request.body)

    @responses.activate
    def test_file_uploader_uploads_part_of_file(self) -> None:
        """Client check and need to upload part of a file."""
        content_length_to_upload, content_range_to_upload = (
            self.setup_partial_file_uploaded(uploaded_size=2)
        )
        self.file_uploader.upload(self.file_path, self.url)

        # Client checked if the file is already upload, the server
        # returned partial-content and the client uploads the missing part
        self.assertEqual(len(responses.calls), 2)

        check_range_to_upload_headers = responses.calls[0].request.headers
        self.assertEqual(
            check_range_to_upload_headers["Content-Range"],
            f"bytes */{len(self.file_content)}",
        )

        file_upload_headers = responses.calls[1].request.headers
        self.assertEqual(
            file_upload_headers["Content-Length"], str(content_length_to_upload)
        )
        self.assertEqual(
            file_upload_headers["Content-Range"], content_range_to_upload
        )

    @responses.activate
    def test_file_upload_file_status_raise_unexpected_status(self) -> None:
        """
        Client check status of the file, server return unexpected code.

        UnexpectedStatusCodeError is raised.
        """
        self.setup_partial_file_uploaded(
            uploaded_size=2, status_code_file_status=requests.codes.teapot
        )

        with self.assertRaises(UnexpectedStatusCodeError):
            self.file_uploader.upload(self.file_path, self.url)

    @responses.activate
    def test_file_uploader_uploads_part_of_file_unexpected_status(self) -> None:
        """
        Client try to upload a file, server return unexpected code.

        UnexpectedStatusCodeError is raised.
        """
        self.setup_partial_file_uploaded(
            uploaded_size=2, status_code_upload=requests.codes.teapot
        )

        with self.assertRaises(UnexpectedStatusCodeError):
            self.file_uploader.upload(self.file_path, self.url)

    @responses.activate
    def test_file_upload_conflict_uploading_file(self) -> None:
        """Client upload file, there is a conflict and uploads it again."""

        def succeeds_after_one_attempt(
            request: requests.PreparedRequest,  # noqa: ARG001, U100
        ) -> tuple[int, dict[str, str], None]:
            first_attempt = getattr(succeeds_after_one_attempt, "first", True)

            status_code: int
            if first_attempt:
                status_code = requests.codes.conflict
            else:
                status_code = requests.codes.ok

            setattr(succeeds_after_one_attempt, "first", False)

            return status_code, {}, None

        uploaded_size = 2
        file_size = len(self.file_content)

        # Missing part to upload
        response_range = f"{uploaded_size}-{file_size}"

        self.add_partial_response(
            requests.codes.partial_content, response_range
        )

        responses.add_callback(
            responses.PUT, self.url, callback=succeeds_after_one_attempt
        )

        self.file_uploader.upload(self.file_path, self.url)

    def add_response(
        self, start_position: int, status: int, /, skip_range: bool = False
    ) -> int:
        """Add a response matching self.file_content (size=chunk size)."""
        end_position = start_position + self.file_uploader._MAX_CHUNK_SIZE_BYTES

        if skip_range:
            headers = {}
        else:
            headers = {"Range": f"bytes={start_position}-{end_position}"}

        responses.add(
            responses.PUT,
            self.url,
            status=status,
            adding_headers=headers,
            match=[
                body_matcher(self.file_content[start_position:end_position]),
            ],
        )
        return min(end_position, len(self.file_content))

    @responses.activate
    def test_upload_server_does_not_include_content_range_response(
        self,
    ) -> None:
        """
        Client upload a chunk, server does not include range header.

        FileUploader.upload raise RuntimeError.
        """
        self.file_uploader._MAX_CHUNK_SIZE_BYTES = 5

        self.add_partial_response(requests.codes.partial_content, "0-0")

        self.add_response(0, requests.codes.ok, skip_range=True)

        with self.assertRaises(RuntimeError):
            self.file_uploader.upload(self.file_path, self.url)

    @responses.activate
    def test_upload_file_in_chunks(self) -> None:
        """Client upload the file in chunks."""
        self.file_uploader._MAX_CHUNK_SIZE_BYTES = 5

        # Nothing has been uploaded yet: from 0 to 0 uploaded
        self.add_partial_response(requests.codes.partial_content, "0-0")

        end_position = self.add_response(0, requests.codes.ok)
        end_position = self.add_response(end_position, requests.codes.ok)
        self.add_response(end_position, requests.codes.created)

        self.file_uploader.upload(self.file_path, self.url)

    def test_parse_range_or_raise_exception_return_range(self) -> None:
        """Return the range (range is in the header and well formatted)."""
        response = mock.create_autospec(
            spec=requests.models.Response, headers={"Range": "bytes=0-20"}
        )

        self.assertEqual(
            FileUploader._parse_range_or_raise_exception(response, 20),
            parse_range_header(response.headers, 20),
        )

    def test_parse_range_or_raise_exception_exception_no_header(self) -> None:
        """Method raise exception: expected header not found."""
        method = "PUT"
        url = "https://test.com"
        request = mock.create_autospec(
            spec=requests.models.Request, method=method, url=url
        )
        response = mock.create_autospec(
            spec=requests.models.Response, request=request, headers={}
        )

        with self.assertRaisesRegex(
            RuntimeError,
            f"^Required range header not in response or invalid for: {method} "
            f"{url} content-range header: None$",
        ):
            FileUploader._parse_range_or_raise_exception(response, 20)

    def test_parse_range_or_raise_exception_exception_invalid_header(
        self,
    ) -> None:
        """Method raise exception: header "range" is not valid."""
        response = mock.create_autospec(
            spec=requests.models.Response,
            request=mock.create_autospec(
                spec=requests.models.Request,
                url="https://example.com",
                method="GET",
            ),
            headers={"Range": "invalid-range"},
        )

        with self.assertRaisesRegex(
            RuntimeError,
            "^Required range header not in response or invalid for: ",
        ):
            FileUploader._parse_range_or_raise_exception(response, 20)

    @responses.activate
    def test_file_upload_server_does_not_return_range(self) -> None:
        """
        Client check status of the file, server return invalid response.

        Server returned "partial content" status code without the Range header.
        """
        responses.add(
            responses.PUT,
            self.url,
            status=requests.codes.partial_content,
            match=[body_matcher(None)],
        )
        with self.assertRaises(RuntimeError):
            self.file_uploader.upload(self.file_path, self.url)

    @responses.activate
    def test_file_upload_conflict_always_conflict(self) -> None:
        """Client upload file, there is always a conflict. Raise exception."""
        self.setup_partial_file_uploaded(
            uploaded_size=2, status_code_upload=requests.codes.conflict
        )

        error_msg = (
            rf"Error uploading .* to {self.url} "
            rf"\({self.file_uploader._MAX_FILE_UPLOAD_RETRIES} retries\)"
        )
        with self.assertRaisesRegex(FileUploadError, error_msg):
            self.file_uploader.upload(self.file_path, self.url)

        url_called_times = self.file_uploader._MAX_FILE_UPLOAD_RETRIES

        # two calls per attempt: to check what needs
        # to be uploaded (server return the range and
        # HTTP 206) and the data upload
        url_called_times *= 2

        responses.assert_call_count(self.url, url_called_times)

    @responses.activate
    def test_file_upload_logs(self) -> None:
        """The uploader logs a message."""
        responses.add(responses.PUT, self.url)
        with self.assertLogsContains(
            f"Uploading {self.file_path} to {self.url}",
            logger=__name__,
            level=logging.INFO,
        ):
            self.file_uploader.upload(self.file_path, self.url)

    @responses.activate
    def test_upload_file_put_fails(self) -> None:
        """FileUpload._upload_file() requests.put() raise exception."""
        responses.add(
            responses.PUT, self.url, body=requests.exceptions.RequestException()
        )

        with (
            self.assertRaises(exceptions.ClientConnectionError),
            open(self.file_path, "rb") as file,
        ):
            self.file_uploader._upload_file(
                file, len(self.file_content), self.url
            )

    @responses.activate
    def test_upload_chunk_retries(self) -> None:
        """FileUploader._upload_chunk() called twice: once failed."""
        responses.add(
            responses.PUT, self.url, body=requests.exceptions.ConnectionError()
        )
        responses.add(responses.PUT, self.url)

        with (
            open(self.file_path, "rb") as file,
            self.spy("requests.put") as mocked_put,
            mock.patch.object(
                self.file_uploader, "_upload_chunk_sleep", autospec=True
            ) as mocked_sleep,
            mmap.mmap(file.fileno(), 3, prot=mmap.PROT_READ) as file_mmaped,
        ):
            file_mmaped.seek(1)
            r = self.file_uploader._upload_chunk(
                url=self.url,
                start_position=1,
                end_position=2,
                file_size=2,
                file=file_mmaped,
            )
        self.assertEqual(mocked_put.call_count, 2)
        self.assertEqual(r.request.headers["Content-Length"], "2")
        self.assertEqual(r.request.headers["Content-Range"], "bytes 1-2/2")
        self.assertEqual(r.request.body, b"es")
        self.assertEqual(mocked_sleep.call_count, 1)
        mocked_sleep.assert_called_with(
            self.file_uploader._CHUNK_UPLOAD_RETRY_WAIT_SECS
        )

    @responses.activate
    def test_retry_block_504(self) -> None:
        # Determine byte-range that we need to upload
        rsp1 = responses.add(
            responses.PUT,
            self.url,
            status=206,
            headers={"Range": "bytes=0-0"},
        )
        # LB Timeout on the main body of the upload
        rsp2 = responses.add(
            responses.PUT,
            self.url,
            status=504,
        )
        # Retry the main upload, successfully
        rsp3 = responses.add(
            responses.PUT,
            self.url,
            status=201,
        )
        self.file_uploader.upload(self.file_path, self.url)
        self.assertEqual(rsp1.call_count, 1)
        self.assertEqual(
            rsp1.calls[0].request.headers.get("Content-Range", None),
            "bytes */12",
        )

        for response in (rsp2, rsp3):
            self.assertEqual(response.call_count, 1)
            self.assertEqual(
                response.calls[0].request.headers.get("Content-Range", None),
                "bytes 0-11/12",
            )

    @responses.activate
    def test_upload_chunk_raise_clientconnectionerror(self) -> None:
        """FileUploader._upload_chunk() called MAX_CHUNK_UPLOAD_RETIRES."""
        responses.add(
            responses.PUT, self.url, body=requests.exceptions.RequestException()
        )

        with (
            self.assertRaises(exceptions.ClientConnectionError),
            open(self.file_path, "rb") as file,
            self.spy("requests.put") as mocked_put,
            mock.patch.object(
                self.file_uploader, "_upload_chunk_sleep", autospec=True
            ) as mocked_sleep,
            mmap.mmap(file.fileno(), 3, prot=mmap.PROT_READ) as file_mmaped,
        ):
            file_mmaped.seek(1)
            self.file_uploader._upload_chunk(
                url=self.url,
                start_position=1,
                end_position=2,
                file_size=2,
                file=file_mmaped,
            )

        self.assertEqual(
            mocked_put.call_count, FileUploader._MAX_CHUNK_UPLOAD_RETRIES
        )
        self.assertEqual(
            mocked_sleep.call_count, FileUploader._MAX_CHUNK_UPLOAD_RETRIES - 1
        )
        mocked_sleep.assert_called_with(
            self.file_uploader._CHUNK_UPLOAD_RETRY_WAIT_SECS
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Rename ``create_experiment_workspace`` to ``create_child_workspace``."""

import copy

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps

from debusine.artifacts.models import TaskTypes


def rename_create_child_workspace_forward(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Rename relevant rows and adjust their data."""
    WorkRequest = apps.get_model("db", "WorkRequest")
    WorkflowTemplate = apps.get_model("db", "WorkflowTemplate")
    for wr in WorkRequest.objects.filter(
        task_type=TaskTypes.SERVER, task_name="create_experiment_workspace"
    ):
        wr.task_name = "create_child_workspace"
        for field in ("task_data", "configured_task_data"):
            if (data := copy.deepcopy(getattr(wr, field))) is not None:
                data["prefix"] = wr.workspace.name
                data["suffix"] = data.pop("experiment_name", None)
                setattr(wr, field, data)
        wr.save()
    for wr in WorkRequest.objects.filter(
        task_type=TaskTypes.WORKFLOW, task_name="create_experiment_workspace"
    ):
        wr.task_name = "create_child_workspace"
        for field in ("task_data", "configured_task_data"):
            if (data := copy.deepcopy(getattr(wr, field))) is not None:
                # The default behaviour if we leave prefix unset is reasonable.
                data["suffix"] = data.pop("experiment_name", None)
                setattr(wr, field, data)
        wr.save()
    for wt in WorkflowTemplate.objects.filter(
        task_name="create_experiment_workspace"
    ):
        wt.task_name = "create_child_workspace"
        for field in ("static_parameters", "runtime_parameters"):
            # Neither of these fields can be None.
            data = copy.deepcopy(getattr(wt, field))
            # prefix is a new field, but we deliberately don't add it to
            # runtime_parameters; the most reasonable default is for it not
            # to be user-configurable.
            if "experiment_name" in data:
                data["suffix"] = data.pop("experiment_name")
            setattr(wt, field, data)
        wt.save()


def rename_create_child_workspace_reverse(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Restore relevant rows to their previous state where possible."""
    WorkRequest = apps.get_model("db", "WorkRequest")
    WorkflowTemplate = apps.get_model("db", "WorkflowTemplate")
    for wr in WorkRequest.objects.filter(
        task_type=TaskTypes.SERVER, task_name="create_child_workspace"
    ):
        wr.task_name = "create_experiment_workspace"
        for field in ("task_data", "configured_task_data"):
            if (data := copy.deepcopy(getattr(wr, field))) is not None:
                data.pop("prefix", None)
                data["experiment_name"] = data.pop("suffix", None)
                setattr(wr, field, data)
        wr.save()
    for wr in WorkRequest.objects.filter(
        task_type=TaskTypes.WORKFLOW, task_name="create_child_workspace"
    ):
        wr.task_name = "create_experiment_workspace"
        for field in ("task_data", "configured_task_data"):
            if (data := copy.deepcopy(getattr(wr, field))) is not None:
                data.pop("prefix", None)
                data["experiment_name"] = data.pop("suffix", None)
                setattr(wr, field, data)
        wr.save()
    for wt in WorkflowTemplate.objects.filter(
        task_name="create_child_workspace"
    ):
        wt.task_name = "create_experiment_workspace"
        for field in ("static_parameters", "runtime_parameters"):
            # Neither of these fields can be None.
            data = copy.deepcopy(getattr(wt, field))
            data.pop("prefix", None)
            if "suffix" in data:
                data["experiment_name"] = data.pop("suffix")
            setattr(wt, field, data)
        wt.save()


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0035_alter_workrequest_status"),
    ]

    operations = [
        migrations.RunPython(
            rename_create_child_workspace_forward,
            rename_create_child_workspace_reverse,
        )
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Tests for the base view code."""

import json
from typing import Any, override

from django.http import HttpResponse
from django.test import Client
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APIRequestFactory

from debusine.db.context import context
from debusine.db.models import Scope, Token, Workspace
from debusine.db.permissioncontext import PermissionContext
from debusine.server.exceptions import DebusineAPIException
from debusine.server.views.base import BaseAPIView, SupersededAPIView
from debusine.test.django import (
    AllowAll,
    BaseDjangoTestCase,
    DenyAll,
    TestCase,
    TestResponseType,
    override_permission,
)

TestResponseType  # Fake use for vulture


class TestBaseAPIView(TestCase):
    """Test the BaseAPIView view."""

    def test_enforce(self) -> None:
        """Test the enforce method."""
        scope = self.playground.get_default_scope()
        user = self.playground.get_default_user()
        context.set_permission_context_scoped(
            scope, PermissionContext.for_ui_user(user)
        )
        view = BaseAPIView()
        with override_permission(Scope, "can_display", AllowAll):
            view.enforce(scope.can_display)
        with (
            override_permission(Scope, "can_display", DenyAll),
            self.assertRaises(DebusineAPIException) as exc,
        ):
            view.enforce(scope.can_display)
        e = exc.exception
        self.assertEqual(
            e.debusine_title, "playground cannot display scope debusine"
        )
        self.assertIsNone(e.debusine_detail)
        self.assertIsNone(e.debusine_validation_errors)
        self.assertEqual(e.debusine_status_code, status.HTTP_403_FORBIDDEN)

    def test_set_current_workspace(self) -> None:
        """Test the set_current_workspace method."""
        workspace: Workspace | str
        error: str | None
        workspace = self.playground.get_default_workspace()
        private_workspace = self.playground.create_workspace(name="private")
        scope = self.playground.get_default_scope()
        user = self.playground.get_default_user()
        context.set_permission_context_scoped(
            scope, PermissionContext.for_ui_user(user)
        )
        view = BaseAPIView()
        for workspace, error in (
            (workspace, None),
            (workspace.name, None),
            (
                private_workspace,
                "Workspace private not found in scope debusine",
            ),
            (
                private_workspace.name,
                "Workspace private not found in scope debusine",
            ),
            (
                "does-not-exist",
                "Workspace does-not-exist not found in scope debusine",
            ),
        ):
            with self.subTest(workspace=repr(workspace)), context.local():
                if not error:
                    view.set_current_workspace(workspace)
                else:
                    with self.assertRaises(DebusineAPIException) as exc:
                        view.set_current_workspace(workspace)
                    self.assertEqual(exc.exception.debusine_detail, error)


class TestWhoami(TestCase):
    """Test authentication using the whoami view."""

    def whoami(
        self, token: Token | str | None = None, client: Client | None = None
    ) -> dict[str, Any]:
        """Call the whoami API view."""
        if client is None:
            client = self.client

        headers = {}
        match token:
            case None:
                pass
            case str():
                headers["token"] = token
            case _:
                headers["token"] = token.key

        response = client.get(reverse("api:whoami"), headers=headers)
        data = getattr(response, "data")
        assert isinstance(data, dict)
        return data

    def test_worker_token(self) -> None:
        """Try a worker token."""
        self.assertEqual(
            self.whoami(self.playground.create_worker_token()),
            {
                "auth": True,
                "user": "<anonymous>",
                "work_request": None,
                "worker": True,
            },
        )

    def test_work_request_token(self) -> None:
        """Try a work request token."""
        token = self.playground.create_work_request_token()
        assert token.work_request is not None
        self.assertEqual(
            self.whoami(token),
            {
                "auth": True,
                "user": "<anonymous>",
                "work_request": token.work_request.id,
                "worker": False,
            },
        )

    def test_user_token(self) -> None:
        """Try a user token."""
        self.assertEqual(
            self.whoami(self.playground.create_user_token()),
            {
                "auth": True,
                "user": "playground",
                "work_request": None,
                "worker": False,
            },
        )

    def test_no_token(self) -> None:
        """Try without a token."""
        self.assertEqual(
            self.whoami(),
            {
                "auth": False,
                "user": "<anonymous>",
                "work_request": None,
                "worker": False,
            },
        )

    def test_session_auth(self) -> None:
        """Try django auth."""
        user = self.playground.get_default_user()
        client = Client()
        client.force_login(user)
        self.assertEqual(
            self.whoami(client=client),
            {
                "auth": False,
                "user": user.username,
                "work_request": None,
                "worker": False,
            },
        )

    def test_token_key_does_not_exist(self) -> None:
        """Try a Token key that does not exist."""
        self.assertEqual(
            self.whoami("does-not-exist"),
            {
                "auth": False,
                "user": "<anonymous>",
                "work_request": None,
                "worker": False,
            },
        )

    def test_token_disabled(self) -> None:
        """Try a disabled Token."""
        # Worker and user token
        token = self.playground.create_user_token(enabled=False)
        self.assertEqual(
            self.whoami(token),
            {
                'auth': False,
                'user': "<anonymous>",
                'worker': False,
                "work_request": None,
            },
        )


class TestSupersededEndpoint(BaseDjangoTestCase):
    """Test the superseded API endpoint."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.factory = APIRequestFactory()

    def request(self, method: str = "GET", **kwargs: str) -> HttpResponse:
        """
        Assert that a method request to view returns an error with detail.

        We're not mounted at URL, so we skip django's test client and just put
        the pieces together ourselves.
        """
        match method:
            case "GET":
                request = self.factory.get("/some-api")
            case "POST":
                request = self.factory.post("/some-api", {})
            case "PATCH":
                request = self.factory.patch("/some-api", {})
            case "DELETE":
                request = self.factory.delete("/some-api")
            case "OPTIONS":
                request = self.factory.options("/some-api")
            case _:  # pragma: no cover
                raise NotImplementedError()
        request._debusine_token = None  # type: ignore[attr-defined]
        response = SupersededAPIView.as_view()(request, **kwargs)
        response.json = lambda: json.loads(response.content)  # type: ignore[attr-defined]
        return response

    def test_unconfigured(self) -> None:
        detail = r"This endpoint is no longer supported\."
        for method in ("GET", "POST", "PATCH", "DELETE"):
            with self.subTest(method=method):
                response = self.request(method=method)
                self.assertResponseProblem(
                    response,
                    title="Superseded API Endpoint",
                    detail_pattern=detail,
                )

    def test_options(self) -> None:
        response = self.request(method="OPTIONS")
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_with_replacement(self) -> None:
        response = self.request(new_endpoint="/new/")
        self.assertResponseProblem(
            response,
            title="Superseded API Endpoint",
            detail_pattern=(
                r"This endpoint is no longer supported\. "
                r"It has been replaced by /new/\."
            ),
        )

    def test_with_replacement_and_version(self) -> None:
        response = self.request(new_endpoint="/new/", client_version="1.0")
        self.assertResponseProblem(
            response,
            title="Superseded API Endpoint",
            detail_pattern=(
                r"This endpoint is no longer supported\. "
                r"It has been replaced by /new/, which is supported by "
                r"debusine-client \(>= 1\.0\)\. Please upgrade your client\."
            ),
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Transaction helpers."""

from typing import Literal, override

import pgtransaction
from pgtransaction import config


class AtomicWithUncachedRetry(pgtransaction.Atomic):
    """
    Like :py:class:`pgtransaction.Atomic`, but without caching ``retry``.

    See https://github.com/AmbitionEng/django-pgtransaction/issues/25.
    """

    @override
    @property
    def retry(self) -> int:
        """
        Lazily load the configured retry value.

        We do this so that atomic decorators can be instantiated without an
        implicit dependency on Django settings being configured.
        """
        if self._retry is not None:  # pragma: no cover
            return self._retry
        retry = config.retry()  # type: ignore[no-untyped-call]
        assert isinstance(retry, int)
        return retry


# Unfortunately we have to copy the complicated type annotations from
# `pgtransaction.atomic` until
# https://github.com/AmbitionEng/django-pgtransaction/issues/25 is fixed.
# This is simplified by removing `using` and `deferrable`, which we don't
# need here.
def atomic(
    savepoint: bool = True,
    durable: bool = False,
    *,
    isolation_level: Literal[
        "READ COMMITTED", "REPEATABLE READ", "SERIALIZABLE"
    ]
    | None = None,
    retry: int | None = None,
    read_mode: Literal["READ WRITE", "READ ONLY"] | None = None,
) -> AtomicWithUncachedRetry:
    """Like :py:func:`pgtransaction.atomic`, but without caching ``retry``."""
    return AtomicWithUncachedRetry(
        using=None,
        savepoint=savepoint,
        durable=durable,
        isolation_level=isolation_level,
        retry=retry,
        read_mode=read_mode,
        deferrable=None,
    )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for signon provider backends."""

from unittest import TestCase

from debusine.server.signon.signon_utils import split_full_name


class SpliyFullName(TestCase):
    """Test split_full_name."""

    def test_empty(self) -> None:
        """Split an empty name."""
        self.assertEqual(split_full_name(""), ("", ""))

    def test_single(self) -> None:
        """Split a single name."""
        self.assertEqual(split_full_name("Test"), ("Test", ""))

    def test_first_last(self) -> None:
        """Split a common First+Last name."""
        self.assertEqual(split_full_name("First Last"), ("First", "Last"))

    def test_middle(self) -> None:
        """Split a 3-part name."""
        self.assertEqual(
            split_full_name("First Middle Last"), ("First Middle", "Last")
        )

    def test_hispanic(self) -> None:
        """Split a 2+2 name as is common in hispanic countries."""
        self.assertEqual(
            split_full_name("First1 First2 Last1 Last2"),
            ("First1 First2", "Last1 Last2"),
        )

    def test_many(self) -> None:
        """Deal gracefully with larger numbers of names."""
        self.assertEqual(
            split_full_name("First1 First2 Last1 Last2 Last3"),
            ("First1 First2", "Last1 Last2 Last3"),
        )

        self.assertEqual(
            split_full_name("First1 First2 First3 Last1 Last2 Last3"),
            ("First1 First2 First3", "Last1 Last2 Last3"),
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test tables."""

from typing import Any, ClassVar, override

import lxml
from django.test import RequestFactory

from debusine.artifacts.models import TaskTypes
from debusine.db.context import context
from debusine.db.models import (
    Group,
    WorkRequest,
    Worker,
    WorkerPool,
    WorkflowTemplate,
)
from debusine.db.models.auth import GroupMembership
from debusine.db.playground import scenarios
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.test.django import TestCase
from debusine.web.views.tables import (
    FilterFailedWorkRequests,
    GroupMembershipAdminTable,
    GroupMembershipTable,
    WorkRequestTable,
    WorkerPoolTable,
    WorkerTable,
    WorkflowPaginator,
    WorkflowTable,
)


class GroupMembershipTableTests(TestCase):
    """Tests for :py:class:`GroupMembershipTable`."""

    scenario = scenarios.DefaultContext(set_current=True)

    @override
    def setUp(self) -> None:
        super().setUp()
        self.request = RequestFactory().get("/")

    def assertUserTableRow(
        self, table: GroupMembershipTable
    ) -> lxml.objectify.ObjectifiedElement:
        tree = self.assertHTMLValid(
            self.render_string(
                "{% load debusine %}{% widget test %}",
                test=table.get_paginator(per_page=10),
                request=self.request,
                user=context.user,
            )
        )
        row = self.assertHasElement(tree, "*/table/tbody/tr")

        self.assertTextContentEqual(
            row.td[0], f"{self.scenario.user.username} (you)"
        )

        self.assertTextContentEqual(
            row.td[1], self.scenario.user.get_full_name()
        )

        return row

    def test_group_without_workspace(self) -> None:
        group = self.playground.create_group("Admins", [self.scenario.user])
        table = GroupMembershipTable(
            self.request, GroupMembership.objects.filter(group=group)
        )
        row = self.assertUserTableRow(table)
        role = row.td[2]
        self.assertFalse(role.xpath("a"))
        self.assertTextContentEqual(role, Group.Roles.MEMBER)
        self.assertEqual(len(row.td), 3)

    def test_admin_group_without_workspace(self) -> None:
        group = self.playground.create_group("Admins", [self.scenario.user])
        gm = group.set_user_role(self.scenario.user, Group.Roles.ADMIN)
        table = GroupMembershipAdminTable(
            self.request, GroupMembership.objects.filter(group=group)
        )
        row = self.assertUserTableRow(table)

        role = row.td[2]
        a = self.assertHasElement(role, "a")
        self.assertEqual(a.get("href"), gm.get_absolute_url_update())
        self.assertTextContentEqual(a, Group.Roles.ADMIN)

        actions = row.td[3]
        a = self.assertHasElement(actions, "a")
        self.assertEqual(a.get("href"), gm.get_absolute_url_remove())
        self.assertTextContentEqual(a, "Remove")

    def test_group_with_workspace(self) -> None:
        group = self.playground.create_group(
            "Admins", [self.scenario.user], workspace=self.scenario.workspace
        )
        table = GroupMembershipTable(
            self.request, GroupMembership.objects.filter(group=group)
        )
        row = self.assertUserTableRow(table)
        role = row.td[2]
        self.assertFalse(role.xpath("a"))
        self.assertTextContentEqual(role, Group.Roles.MEMBER)
        self.assertEqual(len(row.td), 3)

    def test_admin_group_with_workspace(self) -> None:
        group = self.playground.create_group(
            "Admins", [self.scenario.user], workspace=self.scenario.workspace
        )
        gm = group.set_user_role(self.scenario.user, Group.Roles.ADMIN)
        table = GroupMembershipAdminTable(
            self.request, GroupMembership.objects.filter(group=group)
        )
        row = self.assertUserTableRow(table)

        role = row.td[2]
        a = self.assertHasElement(role, "a")
        self.assertEqual(a.get("href"), gm.get_absolute_url_update())
        self.assertTextContentEqual(a, Group.Roles.ADMIN)

        actions = row.td[3]
        a = self.assertHasElement(actions, "a")
        self.assertEqual(a.get("href"), gm.get_absolute_url_remove())
        self.assertTextContentEqual(a, "Remove")


class WorkerTableTests(TestCase):
    """Tests for :py:class:`WorkerTable`."""

    def test_defaults(self) -> None:
        """Test default arguments for custom Pagination object."""
        request = RequestFactory().get("/")
        table = WorkerTable(request, Worker.objects.all())
        self.assertEqual(table.default_order, "name")
        self.assertEqual(
            [c.name for c in table.columns],
            ["type", "name", "pool", "last_seen", "status"],
        )
        self.assertEqual(table.template_name, "web/_worker-list.html")

    def test_worker_pool_options(self) -> None:
        """Test population of worker pool options."""
        pool = self.playground.create_worker_pool(name="testpool")
        self.playground.create_worker(worker_pool=pool)
        request = RequestFactory().get("/")
        table = WorkerTable(request, Worker.objects.all())
        f = table.filters["pool"]
        self.assertEqual(
            list(f.options.keys()),  # type: ignore[attr-defined]
            ["_none", str(pool.pk)],
        )


class WorkerPoolTableTests(TestCase):
    """Tests for :py:class:`WorkerPoolTable`."""

    def test_defaults(self) -> None:
        """Test default arguments for custom Pagination object."""
        request = RequestFactory().get("/")
        table = WorkerPoolTable(request, WorkerPool.objects.all())
        self.assertEqual(table.default_order, "name")
        self.assertEqual(
            [c.name for c in table.columns],
            ["name", "enabled", "instance_wide", "ephemeral", "workers"],
        )


class WorkRequestTableTests(TestCase):
    """Tests for :py:class:`WorkRequestTable`."""

    def test_defaults(self) -> None:
        """Test default arguments for custom Pagination object."""
        request = RequestFactory().get("/")
        table = WorkRequestTable(request, WorkRequest.objects.all())
        self.assertEqual(table.default_order, "-created_at")
        self.assertEqual(
            [c.name for c in table.columns],
            ["id", "created_at", "task_type", "task_name", "status", "result"],
        )


class FilterFailedWorkRequestsTests(TestCase):
    """Tests for :py:class:`FilterFailedWorkRequests`."""

    def test_filter_when_inactive(self) -> None:
        """Test calling filter_queryset with an inactive filter."""
        f = FilterFailedWorkRequests("Test")
        queryset = WorkRequest.objects.all()
        self.assertIs(f.filter_queryset(queryset, None), queryset)


class WorkflowTableTests(TestCase):
    """Tests for :py:class:`WorkflowTable`."""

    template: ClassVar[WorkflowTemplate]
    workflow_1: ClassVar[WorkRequest]
    workflow_2: ClassVar[WorkRequest]

    scenario = scenarios.DefaultContext(set_current=True)

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()

        cls.template = cls.playground.create_workflow_template(
            "name-2", "noop", workspace=cls.scenario.workspace
        )

        cls.workflow_1 = cls.playground.create_workflow(
            cls.template,
            task_data={},
        )
        cls.workflow_2 = cls.playground.create_workflow(
            cls.template,
            task_data={},
        )

    def _table(
        self, query_params: dict[str, Any] | None = None
    ) -> WorkflowTable:
        """Instantiate a table with the given query string parameters."""
        request = RequestFactory().get("/", data=query_params)
        return WorkflowTable(
            request,
            WorkRequest.objects.filter(
                task_type=TaskTypes.WORKFLOW, parent__isnull=True
            ),
        )

    def test_defaults(self) -> None:
        """Test default arguments for custom Pagination object."""
        table = self._table()
        self.assertEqual(table.default_order, "id.desc")
        self.assertEqual(
            [c.name for c in table.columns],
            [
                "id",
                "workflow_template",
                "status",
                "result",
                "started_at",
                "completed_at",
                "wr_count",
                "last_activity",
                "started_by",
            ],
        )

    def test_filter_form_with_issues(self) -> None:
        """Client submit filter form which is not valid."""
        table = self._table(
            {
                "filter-statuses": "does-not-exist",
                "filter-with_failed_work_requests": "1",
            }
        )
        f = table.filters["statuses"]
        self.assertIsNone(f.value)
        f = table.filters["with_failed_work_requests"]
        self.assertEqual(f.value, "1")

    def assert_filtered(self, workflow: WorkRequest, **kwargs: Any) -> None:
        """Make a request with the form and assert one workflow is listed."""
        table = self._table({f"filter-{k}": v for k, v in kwargs.items()})
        self.assertQuerySetEqual(table.rows, [workflow])

    def test_list_filtering_started_by(self) -> None:
        """View detail filters by started_by user."""
        user = self.playground.create_user(username="relevant")
        self.workflow_1.created_by = user
        self.workflow_1.save()
        self.assert_filtered(self.workflow_1, started_by=user.username)

    def test_list_filtering_status(self) -> None:
        """View detail filters by status."""
        self.workflow_1.status = WorkRequest.Statuses.COMPLETED
        self.workflow_1.save()
        self.assert_filtered(
            self.workflow_1, statuses=[WorkRequest.Statuses.COMPLETED]
        )

    def test_list_filtering_runtime_status(self) -> None:
        """View detail filters by runtime status."""
        self.workflow_1.status = WorkRequest.Statuses.RUNNING
        self.workflow_1.save()
        self.workflow_2.status = WorkRequest.Statuses.RUNNING
        self.workflow_2.save()
        self.workflow_1.create_child_worker("noop")
        self.assert_filtered(self.workflow_1, statuses=["running__blocked"])

    def test_list_filtering_runtime_any(self) -> None:
        """
        Test Workflow with running status is displayed.

        No need for any specific WorkRequest.workflow_running_status.
        """
        self.workflow_1.status = WorkRequest.Statuses.RUNNING
        self.workflow_1.save()
        self.workflow_1.create_child_worker("noop")
        self.assert_filtered(self.workflow_1, statuses=["running__any"])

    def test_list_filtering_result(self) -> None:
        """View detail filters by result."""
        self.workflow_1.result = WorkRequest.Results.SUCCESS
        self.workflow_1.save()
        self.assert_filtered(
            self.workflow_1, results=[WorkRequest.Results.SUCCESS]
        )

    def test_list_filtering_with_failed_work_requests(self) -> None:
        """View detail filters by with failed work requests."""
        table = self._table({"filter-with_failed_work_requests": "1"})
        self.assertQuerySetEqual(table.rows, [])

        wr = self.workflow_1.create_child_worker("noop")
        wr.result = WorkRequest.Results.FAILURE
        wr.save()
        self.assert_filtered(self.workflow_1, with_failed_work_requests="1")

        for _ in range(2):
            wr = self.workflow_2.create_child_worker("noop")
            wr.result = WorkRequest.Results.FAILURE
            wr.save()
        table = self._table({"filter-with_failed_work_requests": "1"})
        self.assertQuerySetEqual(
            table.rows, [self.workflow_1, self.workflow_2], ordered=False
        )

    def test_list_filtering_with_failed_work_requests_allow_failure(
        self,
    ) -> None:
        """Workflows with only allow_failure failures are excluded."""
        # workflow_1 has a failure with allow_failure=True: should be excluded
        wr_allow = self.workflow_1.create_child_worker(
            "noop",
            workflow_data=WorkRequestWorkflowData(
                allow_failure=True, display_name="test", step="test"
            ),
        )
        wr_allow.result = WorkRequest.Results.FAILURE
        wr_allow.save()

        # workflow_2 has a failure without allow_failure: should be included
        wr_no_allow = self.workflow_2.create_child_worker("noop")
        wr_no_allow.result = WorkRequest.Results.FAILURE
        wr_no_allow.save()

        self.assert_filtered(self.workflow_2, with_failed_work_requests="1")

    def test_paginator_allow_failure_skips_result_count(self) -> None:
        """Failed work requests with allow_failure are not counted."""
        # Child with allow_failure=True: should be excluded from result counts
        wr_allow = self.workflow_1.create_child_worker(
            "noop",
            workflow_data=WorkRequestWorkflowData(
                allow_failure=True, display_name="test", step="allow"
            ),
        )
        wr_allow.result = WorkRequest.Results.FAILURE
        wr_allow.save()

        # Child without allow_failure: should be counted as a failure
        wr_no_allow = self.workflow_1.create_child_worker(
            "noop",
            workflow_data=WorkRequestWorkflowData(
                display_name="test", step="no-allow"
            ),
        )
        wr_no_allow.result = WorkRequest.Results.FAILURE
        wr_no_allow.save()

        table = self._table()
        paginator = table.get_paginator(per_page=10)
        ctx = paginator.get_context_data()

        self.assertIsInstance(paginator, WorkflowPaginator)
        # Full equality test: only the non-allow_failure child is counted
        self.assertEqual(
            ctx["work_request_result_counts"],
            {self.workflow_1.pk: {WorkRequest.Results.FAILURE: 1}},
        )

    def test_list_filtering_workflow_templates(self) -> None:
        """View detail filters by workflow_templates."""
        other_template = self.playground.create_workflow_template(
            "other_template", "noop"
        )
        self.workflow_2.workflow_template = other_template
        self.workflow_2.save()
        self.assert_filtered(
            self.workflow_1, workflow_templates=self.template.name
        )

    def test_sorting_id(self) -> None:
        """Test output of workflows for "id" and "-id" sorting."""
        table = self._table({"order": "id.desc"})
        self.assertQuerySetEqual(table.rows, [self.workflow_2, self.workflow_1])
        table = self._table({"order": "id.asc"})
        self.assertQuerySetEqual(table.rows, [self.workflow_1, self.workflow_2])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Migrate existing Workflow Templates to use runtime_parameters."""

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps

from debusine.client.models import RuntimeParameter

# This migration needs a snapshot of the available parameters for each
# workflow at the relevant point in Debusine's history, and Django's normal
# migration system doesn't keep track of that information.  The following is
# the result of checking out 51b9140e41d6a4677e349c4be6eed0ab06b4bbd6 (when
# this migration initially landed on devel) and running:
#
#   from debusine.artifacts.models import TaskTypes
#   from debusine.tasks import BaseTask
#
#   print(
#       {
#           task_name: sorted(workflow_cls.task_data_type.__fields__)
#           for task_name, workflow_cls in BaseTask._sub_tasks[
#               TaskTypes.WORKFLOW
#           ].items()
#       }
#   )

workflow_parameters = {
    "autopkgtest": [
        "arch_all_build_architecture",
        "architectures",
        "backend",
        "binary_artifacts",
        "codename",
        "context_artifacts",
        "debug_level",
        "enable_regression_tracking",
        "exclude_tests",
        "extra_environment",
        "extra_repositories",
        "fail_on",
        "include_tests",
        "needs_internet",
        "prefix",
        "qa_suite",
        "reference_prefix",
        "reference_qa_results",
        "source_artifact",
        "task_configuration",
        "timeout",
        "update_qa_results",
        "vendor",
    ],
    "blhc": [
        "arch_all_build_architecture",
        "codename",
        "enable_regression_tracking",
        "extra_flags",
        "package_build_logs",
        "prefix",
        "qa_suite",
        "reference_prefix",
        "reference_qa_results",
        "source_artifact",
        "task_configuration",
        "update_qa_results",
        "vendor",
    ],
    "create_experiment_workspace": [
        "experiment_name",
        "expiration_delay",
        "owner_group",
        "public",
        "task_configuration",
        "workflow_template_names",
    ],
    "debdiff": [
        "arch_all_build_architecture",
        "binary_artifacts",
        "codename",
        "extra_flags",
        "original",
        "source_artifact",
        "task_configuration",
        "vendor",
    ],
    "debian_pipeline": [
        "arch_all_build_architecture",
        "architectures",
        "architectures_allowlist",
        "architectures_denylist",
        "autopkgtest_backend",
        "codename",
        "enable_autopkgtest",
        "enable_blhc",
        "enable_confirmation",
        "enable_debdiff",
        "enable_lintian",
        "enable_make_signed_source",
        "enable_piuparts",
        "enable_regression_tracking",
        "enable_reverse_dependencies_autopkgtest",
        "enable_upload",
        "extra_repositories",
        "lintian_backend",
        "lintian_fail_on_severity",
        "make_signed_source_key",
        "make_signed_source_purpose",
        "piuparts_backend",
        "piuparts_environment",
        "publish_replace",
        "qa_failure_policy",
        "qa_suite",
        "regression_tracking_qa_results",
        "sbuild_backend",
        "sbuild_environment_variant",
        "signing_template_names",
        "source_artifact",
        "suite",
        "task_configuration",
        "upload_binary_key",
        "upload_delayed_days",
        "upload_include_binaries",
        "upload_include_source",
        "upload_key",
        "upload_merge_uploads",
        "upload_require_signature",
        "upload_since_version",
        "upload_target",
        "upload_target_distribution",
        "vendor",
    ],
    "lintian": [
        "arch_all_build_architecture",
        "architectures",
        "backend",
        "binary_artifacts",
        "codename",
        "enable_regression_tracking",
        "exclude_tags",
        "fail_on_severity",
        "include_tags",
        "output",
        "prefix",
        "qa_suite",
        "reference_prefix",
        "reference_qa_results",
        "source_artifact",
        "task_configuration",
        "update_qa_results",
        "vendor",
    ],
    "make_signed_source": [
        "architectures",
        "binary_artifacts",
        "codename",
        "key",
        "prefix",
        "purpose",
        "sbuild_backend",
        "signing_template_artifacts",
        "task_configuration",
        "vendor",
    ],
    "noop": ["task_configuration"],
    "package_publish": [
        "binary_artifacts",
        "replace",
        "source_artifact",
        "suite_variables",
        "target_suite",
        "task_configuration",
        "unembargo",
        "update_indexes",
    ],
    "package_upload": [
        "arch_all_build_architecture",
        "binary_artifacts",
        "binary_key",
        "codename",
        "delayed_days",
        "key",
        "merge_uploads",
        "prefix",
        "require_signature",
        "since_version",
        "source_artifact",
        "target",
        "target_distribution",
        "task_configuration",
        "vendor",
    ],
    "piuparts": [
        "arch_all_build_architecture",
        "architectures",
        "backend",
        "binary_artifacts",
        "codename",
        "enable_regression_tracking",
        "environment",
        "extra_repositories",
        "prefix",
        "qa_suite",
        "reference_prefix",
        "reference_qa_results",
        "source_artifact",
        "task_configuration",
        "update_qa_results",
        "vendor",
    ],
    "qa": [
        "arch_all_build_architecture",
        "architectures",
        "architectures_allowlist",
        "architectures_denylist",
        "autopkgtest_backend",
        "binary_artifacts",
        "codename",
        "enable_autopkgtest",
        "enable_blhc",
        "enable_debdiff",
        "enable_lintian",
        "enable_piuparts",
        "enable_regression_tracking",
        "enable_reverse_dependencies_autopkgtest",
        "extra_repositories",
        "fail_on",
        "lintian_backend",
        "lintian_fail_on_severity",
        "package_build_logs",
        "piuparts_backend",
        "piuparts_environment",
        "prefix",
        "qa_suite",
        "reference_prefix",
        "reference_qa_results",
        "source_artifact",
        "task_configuration",
        "update_qa_results",
        "vendor",
    ],
    "reverse_dependencies_autopkgtest": [
        "arch_all_build_architecture",
        "architectures",
        "backend",
        "binary_artifacts",
        "codename",
        "context_artifacts",
        "debug_level",
        "enable_regression_tracking",
        "extra_repositories",
        "packages_allowlist",
        "packages_denylist",
        "prefix",
        "qa_suite",
        "reference_prefix",
        "reference_qa_results",
        "source_artifact",
        "task_configuration",
        "update_qa_results",
        "vendor",
    ],
    "sbuild": [
        "arch_all_build_architecture",
        "architectures",
        "backend",
        "binnmu",
        "build_profiles",
        "environment_variant",
        "extra_repositories",
        "input",
        "prefix",
        "retry_delays",
        "signing_template_names",
        "target_distribution",
        "task_configuration",
    ],
    "update_environments": ["targets", "task_configuration", "vendor"],
    "update_suite": ["suite_collection", "task_configuration"],
    "update_suites": [
        "force_basic_indexes",
        "only_suites",
        "task_configuration",
    ],
}


def set_workflowtemplate_runtime_parameters(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Migrate existing Workflow Templates to use runtime_parameters."""
    WorkflowTemplate = apps.get_model("db", "WorkflowTemplate")

    for workflow_template in WorkflowTemplate.objects.all():
        for key in workflow_parameters.get(workflow_template.task_name, []):
            if key not in workflow_template.static_parameters:
                workflow_template.runtime_parameters[key] = RuntimeParameter.ANY
        workflow_template.save()


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0033_workflowtemplate_add_runtime_parameters"),
    ]

    operations = [
        migrations.RunPython(
            set_workflowtemplate_runtime_parameters, migrations.RunPython.noop
        ),
    ]



# Generated by Django 4.2.21 on 2025-09-18 16:26

from django.db import migrations, models

import debusine.db.models.work_requests


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0018_alter_assetusagerole_role'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='workflowtemplate',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(('task_name', 'update_suites'), _negated=True),
                    (
                        'name',
                        debusine.db.models.work_requests.WellKnownWorkflowTemplate[
                            'UPDATE_SUITES'
                        ],
                    ),
                    _connector='OR',
                ),
                name='db_workflowtemplate_update_suites',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Base infrastructure for web views."""

import abc
import logging
from collections.abc import Callable
from typing import Any, NoReturn, TYPE_CHECKING, TypeAlias, override

from django.core.exceptions import PermissionDenied
from django.http import Http404, HttpRequest
from django.http.response import HttpResponseBase
from django.template.context import Context
from django.views.generic import (
    CreateView,
    DeleteView,
    DetailView,
    ListView,
    UpdateView,
)
from django.views.generic.base import ContextMixin, View
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import FormMixin, FormView
from django.views.generic.list import MultipleObjectMixin

from debusine.artifacts.models import CollectionCategory
from debusine.db.context import ContextConsistencyError, context
from debusine.db.models import Scope, Workspace
from debusine.db.models.permissions import format_permission_check_error
from debusine.db.permissioncontext import PermissionContext
from debusine.utils import debusine_version
from debusine.web.views.places import IndexPlace, Place

if TYPE_CHECKING:
    CreateViewBase: TypeAlias = CreateView
    DeleteViewBase: TypeAlias = DeleteView
    DetailViewBase: TypeAlias = DetailView
    ListViewBase: TypeAlias = ListView
    UpdateViewBase: TypeAlias = UpdateView
    FormMixinBase: TypeAlias = FormMixin
    FormViewBase: TypeAlias = FormView
    SingleObjectMixinBase: TypeAlias = SingleObjectMixin
    MultipleObjectMixinBase: TypeAlias = MultipleObjectMixin
else:
    # Django's generic views don't support generic types at run-time yet.
    from debusine.utils.typing_utils import patch_runtime_generic_support

    CreateViewBase = patch_runtime_generic_support(CreateView)
    DeleteViewBase = patch_runtime_generic_support(DeleteView)
    DetailViewBase = patch_runtime_generic_support(DetailView)
    ListViewBase = patch_runtime_generic_support(ListView)
    UpdateViewBase = patch_runtime_generic_support(UpdateView)
    FormMixinBase = patch_runtime_generic_support(FormMixin)
    FormViewBase = patch_runtime_generic_support(FormView)
    SingleObjectMixinBase = patch_runtime_generic_support(SingleObjectMixin)
    MultipleObjectMixinBase = patch_runtime_generic_support(MultipleObjectMixin)


class Widget(abc.ABC):
    """Base class for template-renderable elements."""

    @abc.abstractmethod
    def render(self, context: Context) -> str:
        """Render the element."""


class TemplateWidget(Widget):
    """Widget implementation that uses a template."""

    template_name: str

    def get_template_name(self) -> str:
        """Return the name of the template to use."""
        return self.template_name

    def get_context_data(self) -> dict[str, Any]:
        """Return the template context."""
        return {"widget": self}

    @override
    def render(self, context: Context) -> str:
        assert context.template is not None
        template = context.template.engine.get_template(
            self.get_template_name()
        )
        with context.update(self.get_context_data()):
            return template.render(context)


class BaseUIView(ContextMixin, View):
    """Base class for Debusine web views."""

    base_template = "web/_base_scoped.html"
    title = ""

    allow_work_request_tokens: bool = False

    # If a member is defined for these methods
    http_init_view_method_names = [
        "get",
        "post",
        "put",
        "patch",
        "delete",
        "head",
    ]

    @override
    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        """Call self.init_view when appropriate."""
        try:
            pc = context.pc
            work_request_token = pc.work_request_token
            non_permission_bearing_token = pc.is_non_permission_bearing_token
        except ContextConsistencyError:
            work_request_token = None
            non_permission_bearing_token = False

        if work_request_token and not self.allow_work_request_tokens:
            raise PermissionDenied(
                "work request tokens cannot be used with web views"
            )
        if non_permission_bearing_token:
            raise PermissionDenied(
                "special purpose tokens cannot be used with web views"
            )

        assert request.method is not None
        if (
            method := request.method.lower()
        ) in self.http_init_view_method_names:
            if hasattr(self, method):
                self.init_view()
        return super().dispatch(request, *args, **kwargs)

    def init_view(self) -> None:
        """
        Initialize the view.

        Call this method to lookup common objects, or perform permission
        checks.
        """

    def get_place(self) -> Place:
        """Get a Place object for this page."""
        return IndexPlace(title=self.title, url=self.request.get_full_path())

    def get_base_template(self) -> str:
        """Return the name of the base template to use with this view."""
        return self.base_template

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """
        Add base template information to the template context.

        Added elements:

        * base_template: name of the base template to load
        * title: string to use as default page title and header
        """
        ctx = super().get_context_data(**kwargs)
        ctx["theme"] = self.request.COOKIES.get("theme", None)
        ctx["base_template"] = self.get_base_template()
        ctx["place"] = self.get_place()
        ctx["scopes"] = Scope.objects.can_display(context.pc).order_by("name")
        if context.workspace:
            ctx["other_workspaces"] = list(
                Workspace.objects.can_display(context.pc)
                .exclude(pk=context.workspace.pk)
                .order_by("name")
            )
            ctx["workspace_collections"] = (
                context.workspace.collections.exclude(
                    category=CollectionCategory.WORKFLOW_INTERNAL
                )
                .can_display(context.pc)
                .order_by("name", "category")
            )
            ctx["workflow_templates"] = list(
                context.workspace.workflowtemplate_set.can_display(
                    context.pc
                ).order_by("name")
            )
        ctx["debusine_version"] = debusine_version()
        return ctx

    def raise_workspace_not_found(self, workspace: str | Workspace) -> NoReturn:
        """
        Raise Http404 for a workspace not found.

        This is abstracted to provide consistent error messages also in case of
        permission denied, to prevent leaking the presence of workspaces in
        unaccessible scopes
        """
        if isinstance(workspace, Workspace):
            workspace = workspace.name
        raise Http404(
            f"Workspace {workspace} not found in scope {context.scope}, or "
            f"you are not authorized to see it"
        )

    def set_current_workspace(
        self, workspace: str | Workspace, *, check_visibility: bool = True
    ) -> None:
        """Set the current workspace in context."""
        if isinstance(workspace, str):
            try:
                workspace = Workspace.objects.get_for_context(name=workspace)
            except Workspace.DoesNotExist:
                self.raise_workspace_not_found(workspace)
        try:
            workspace.set_current(check_visibility=check_visibility)
        except ContextConsistencyError:
            # Turn exception in a 404 response.
            # 404 is used instead of 403 as an attempt to prevent leaking which
            # private workspace exists that the user cannot see
            logging.debug("permission denied on %s reported as 404", workspace)
            self.raise_workspace_not_found(workspace)

    def enforce(self, predicate: Callable[[PermissionContext], bool]) -> None:
        """Enforce a permission predicate."""
        if predicate(context.pc):
            return

        raise PermissionDenied(
            format_permission_check_error(predicate, context.pc),
        )


class WorkspaceView(BaseUIView, View):
    """Common structure for views under the /<workspace>/ URL namespace."""

    @override
    def init_view(self) -> None:
        """Set the current workspace."""
        super().init_view()
        self.set_current_workspace(self.kwargs["wname"])
        assert context.workspace is not None
        self.enforce(context.workspace.can_display)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine WorkRequest view."""

import abc
import datetime as dt
import functools
from dataclasses import dataclass
from typing import Any, TYPE_CHECKING, cast, override

from django.contrib import messages
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect, render
from django.template.defaultfilters import filesizeformat
from django.utils.duration import duration_string
from django.views.generic.base import View
from rest_framework import status

import debusine.worker.tags as wtags
from debusine.artifacts.models import TaskTypes
from debusine.db.context import context
from debusine.db.models import Artifact, User, WorkRequest
from debusine.db.models.work_requests import (
    CannotAbort,
    CannotRetry,
    CannotUnblock,
    WorkRequestQuerySet,
)
from debusine.db.tasks import DBTask
from debusine.server.tasks.wait import Confirm
from debusine.server.workflows import Workflow
from debusine.server.workflows.models import WorkRequestManualUnblockAction
from debusine.tasks.models import LookupMultiple, LookupSingle, OutputData
from debusine.web.forms import WorkRequestConfirmForm, WorkRequestUnblockForm
from debusine.web.views import sidebar
from debusine.web.views.base import (
    DetailViewBase,
    FormMixinBase,
    ListViewBase,
    SingleObjectMixinBase,
    TemplateWidget,
    WorkspaceView,
)
from debusine.web.views.http_errors import HttpError400, catch_http_errors
from debusine.web.views.input_fields import TaskInputsWidget
from debusine.web.views.mixins import RightbarUIMixin
from debusine.web.views.places import Place
from debusine.web.views.table import TableMixin
from debusine.web.views.tables import ArtifactTable, WorkRequestTable
from debusine.web.views.ui.work_requests import WorkRequestUI
from debusine.web.views.view_utils import (
    highlight_as_json,
    highlight_as_yaml,
    syntax_highlight,
)

if TYPE_CHECKING:
    from debusine.db.models.artifacts import ArtifactWithComplete

    ArtifactWithComplete  # fake usage for vulture


class Tab(TemplateWidget):
    """Define a tab."""

    def __init__(
        self, name: str, label: str, context: dict[str, Any] | None = None
    ) -> None:
        """Store name and label."""
        self.name = name
        self.label = label
        # TODO: this is here only to ease transition from previous WorkRequest
        # plugins. Eventually plugins will move their get_context_data to a Tab
        # subclass
        self.context = context or {}

    @override
    def get_template_name(self) -> str:
        return f"web/work_request-detail-{self.name}.html"

    @override
    def get_context_data(self) -> dict[str, Any]:
        ctx = super().get_context_data()
        ctx.update(self.context)
        return ctx


class WorkRequestObjectMixin(WorkspaceView):
    """
    Add permission filtering to a Single/MultipleObjectMixin.

    Since this inherits from WorkspaceView, this also sets the workspace in the
    context and checks that it is accessible.
    """

    def get_queryset(self) -> WorkRequestQuerySet[WorkRequest]:
        """Filter work requests by current workspace."""
        # Both SingleObjectMixin and MultipleObjectsMixin define a get_queryset
        # without a common base, so we cannot inherit from a base class with
        # the right signature.
        qs = cast(
            WorkRequestQuerySet[WorkRequest],
            super().get_queryset(),  # type: ignore[misc]
        )
        return qs.in_current_workspace().can_display(context.pc)


@dataclass(frozen=True)
class InputArtifactsRow:
    """Represents a InputArtifact row for the template."""

    input_label: str
    lookup_display: LookupSingle | LookupMultiple
    artifacts: list[Artifact]


class WorkRequestDetailView(
    RightbarUIMixin,
    WorkRequestObjectMixin,
    DetailViewBase[WorkRequest],
):
    """Show a work request."""

    queryset = WorkRequest.objects.select_related(
        "workspace__scope", "workflow_root"
    )
    context_object_name = "work_request"
    template_name = "web/work_request-detail.html"

    @functools.cached_property
    def _validation_error_message(self) -> str | None:
        """If the work request fails validation, return a suitable message."""
        try:
            self.object.full_clean()
        except ValidationError as e:
            return str(e)
        else:
            return None

    @functools.cached_property
    def ui(self) -> WorkRequestUI:
        """Return the UI helper for this work request."""
        return self.object.ui(self.request)

    @functools.cached_property
    def task(self) -> DBTask[Any] | None:
        """
        Get the task for the current work request.

        :return: the task, or None if the task data fails to validate
        """
        return self.ui.task

    @functools.cached_property
    def plugin(self) -> "WorkRequestPlugin | None":
        """Return the work request plugin, if defined."""
        if self.task is None:
            return None
        return WorkRequestPlugin.plugin_for(self.task)

    @override
    def get_sidebar_items(self) -> list[sidebar.SidebarItem]:
        items = super().get_sidebar_items()
        items.append(self.ui.sidebar_work_request())
        if hasattr(self.object, "superseded"):
            items.append(self.ui.sidebar_work_request_superseded())
        if self.object.supersedes:
            items.append(self.ui.sidebar_work_request_supersedes())
        items.append(self.ui.sidebar_workflow())
        items.append(self.ui.sidebar_work_request_status())
        items.append(sidebar.create_workspace(self.object.workspace))
        items.append(
            sidebar.create_user(self.object.created_by, context=self.object)
        )
        items.append(sidebar.create_created_at(self.object.created_at))
        items.append(sidebar.create_worker(self.object.worker))
        items.append(self.ui.sidebar_work_request_started_at())
        items.append(self.ui.sidebar_work_request_duration())
        items.append(sidebar.create_expire_at(self.object.expire_at))
        return items

    @override
    def get_place(self) -> Place:
        return self.ui.place

    def get_manual_unblock_log(self) -> list[dict[str, Any]]:
        """Get a representation of the manual unblock log."""
        workflow_data = self.object.workflow_data
        if workflow_data.manual_unblock is None:
            return []

        users = {
            user.id: user
            for user in get_user_model().objects.filter(
                id__in={log.user_id for log in workflow_data.manual_unblock.log}
            )
        }
        return [
            {
                "timestamp": log.timestamp,
                "user": users.get(log.user_id, f"Deleted user {log.user_id}"),
                "action": log.action,
                "notes": log.notes,
            }
            for log in workflow_data.manual_unblock.log
        ]

    def get_default_description(self) -> str | None:
        """Return default description for the Task."""
        if self.task:
            return self.task.get_description()

        try:
            return DBTask.class_from_name(
                TaskTypes(self.object.task_type), self.object.task_name
            ).__doc__
        except ValueError:
            return None

    def get_tabs(self) -> list[Tab]:
        """Return the tabs to show in the view."""
        tabs: list[Tab] = [
            Tab(name="work_request", label="Work request"),
        ]
        if not self.object.is_workflow or self.object.output_data:
            tabs += [
                Tab(name="output", label="Output"),
            ]
        tabs += [
            Tab(name="dependencies", label="Dependencies"),
            Tab(name="internals", label="Internals"),
        ]

        # After the initial tab list is set, allow the plugin to add to it
        if self.plugin is not None:
            self.plugin.add_tabs(tabs)
        return tabs

    @staticmethod
    def _get_workflow_input_artifacts(
        task: Workflow[Any, Any],
    ) -> list[InputArtifactsRow]:
        """
        Build context data describing the input artifacts of a workflow.

        Return tuple:
        * list of InputArtifactsRows (containing information for rendering)
        * list of all referenced Artifact objects
        """
        input_artifacts = task.get_input_artifacts()

        # Artifact ids for each InputArtifact
        input_artifacts_artifact_ids: list[list[int]] = []

        all_artifact_ids: set[int] = set()

        # Collect all the artifact_ids
        for input_artifact in input_artifacts:
            artifact_ids = input_artifact.resolved_ids()
            input_artifacts_artifact_ids.append(artifact_ids)
            all_artifact_ids.update(artifact_ids)

        # Query all the artifacts
        artifacts_qs = (
            Artifact.objects.filter(pk__in=all_artifact_ids)
            .order_by("pk")
            .annotate_complete()
        )

        all_artifacts = list(artifacts_qs)
        artifacts_by_id: dict[int, Artifact] = {a.id: a for a in all_artifacts}

        input_artifact_rows: list[InputArtifactsRow] = []

        for input_artifact, artifact_ids in zip(
            input_artifacts, input_artifacts_artifact_ids, strict=True
        ):
            artifacts_for_row = [
                artifacts_by_id[artifact_id]
                for artifact_id in sorted(set(artifact_ids))
                if artifact_id in artifacts_by_id
            ]

            match input_artifact.lookup:
                case str() | int():
                    lookup_display = syntax_highlight(
                        "text", str(input_artifact.lookup)
                    )
                case _:
                    lookup_display = highlight_as_yaml(
                        input_artifact.lookup.export()
                    )

            input_artifact_rows.append(
                InputArtifactsRow(
                    input_label=input_artifact.label,
                    lookup_display=lookup_display,
                    artifacts=artifacts_for_row,
                )
            )

        return input_artifact_rows

    @staticmethod
    def _get_workflow_parents(work_request: WorkRequest) -> list[WorkRequest]:
        workflow_parents = []

        parent = work_request.parent
        while parent is not None:
            workflow_parents.append(parent)
            parent = parent.parent

        return list(reversed(workflow_parents))

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """Return context for this view's tab."""
        ctx = super().get_context_data(**kwargs)

        ctx["tabs"] = self.get_tabs()

        ctx["description_template"] = (
            "web/_work_request-generic-description.html"
        )
        ctx["description_data"] = {
            "default_description": self.get_default_description()
        }

        ctx["manual_unblock_log_entries"] = self.get_manual_unblock_log()
        ctx["manual_unblock_form"] = WorkRequestUnblockForm()

        if self._validation_error_message is not None:
            ctx["validation_error"] = self._validation_error_message
        elif self.plugin:
            # plugin_view for WorkRequest.task_name exists, but the
            # response will return the generic WorkRequest view
            ctx["specialized_view_path"] = self.request.path

        if self.object.requires_confirmation:
            assert isinstance(self.task, Confirm)
            ctx["confirm_form"] = WorkRequestConfirmForm()
            ctx["confirm_label"] = self.task.data.confirm_label
            ctx["deny_label"] = self.task.data.deny_label

        if isinstance(self.task, Workflow):
            ctx["input_artifacts_list"] = self._get_workflow_input_artifacts(
                self.task
            )
        elif self.task is not None:
            ctx["input_fields"] = TaskInputsWidget(self.task)
            ctx["input_fields_internal"] = TaskInputsWidget(
                self.task,
                internal=True,
                element_id="task-input-fields-internal",
            )

        built_artifacts_table = ArtifactTable(
            self.request,
            Artifact.objects.filter(
                created_by_work_request=self.object
            ).annotate_complete(),
            prefix="built_artifacts",
            fragment="output",
        )
        built_artifacts_table_paginator = built_artifacts_table.get_paginator(
            per_page=20
        )
        ctx["built_artifacts_table"] = built_artifacts_table_paginator

        ctx["task_data"] = highlight_as_yaml(self.object.task_data)

        if (dynamic_task_data := self.object.dynamic_task_data) is not None:
            dynamic_task_data = dynamic_task_data.copy()
            task_input_fields = dynamic_task_data.pop("task_input_fields", None)
            ctx["show_task_data_dynamic"] = True
            ctx["task_data_dynamic"] = highlight_as_yaml(dynamic_task_data)
            if self.task is None:
                ctx["task_input_fields"] = highlight_as_yaml(task_input_fields)
                ctx["task_input_fields_raw"] = True
        else:
            ctx["show_task_data_dynamic"] = False

        if (
            configured_task_data := self.object.configured_task_data
        ) is not None:
            ctx["show_task_data_configured"] = True
            ctx["task_data_configured"] = highlight_as_yaml(
                configured_task_data
            )
            ctx["scheduler_tags"] = highlight_as_yaml(
                {
                    "provided": sorted(self.object.scheduler_tags_provided),
                    "required": sorted(self.object.scheduler_tags_required),
                }
            )
        else:
            ctx["show_task_data_configured"] = False

        ctx["work_request_output_data"] = self._format_output_data(
            self.object.output_data
        )
        ctx["work_request_event_reactions_json"] = highlight_as_json(
            self.object.event_reactions_json
        )
        ctx["work_request_output_data_json"] = highlight_as_json(
            self.object.output_data_json
        )
        ctx["work_request_workflow_data_json"] = highlight_as_json(
            self.object.workflow_data_json
        )

        ctx["work_request_dependencies"] = WorkRequestTable(
            self.request, self.object.dependencies.all(), prefix="dependencies"
        ).get_paginator(per_page=10)
        ctx["work_request_reverse_dependencies"] = WorkRequestTable(
            self.request,
            self.object.reverse_dependencies.all(),
            prefix="reverse_dependencies",
        ).get_paginator(per_page=10)

        ctx["workflow_parents"] = self._get_workflow_parents(self.object)

        # Let the plugin override parts of the context, if needed
        if self.plugin:
            ctx.update(self.plugin.get_context_data())

        return ctx

    @staticmethod
    def _format_output_data(
        output_data: OutputData | None,
    ) -> dict[str, Any]:
        """
        Format OutputData into a dictionary with human-readable values.

        - Memory and disk sizes are formatted as MB/GB.
        - Durations are shown in hours, minutes, and seconds.
        """
        if output_data is None:
            return {}

        runtime_statistics = {}

        if output_data.runtime_statistics is not None:
            for key, value in output_data.runtime_statistics:
                if (
                    key
                    in (
                        "memory",
                        "available_memory",
                        "disk_space",
                        "available_disk_space",
                    )
                    and value is not None
                ):
                    value = filesizeformat(value)

                if key in ("duration", "cpu_time") and value is not None:
                    value = duration_string(dt.timedelta(seconds=value))

                key = key.replace("_", " ").capitalize().replace("Cpu ", "CPU ")

                runtime_statistics[key] = value

        errors: list[dict[str, str]] = []
        if output_data.errors is not None:
            errors += [
                {"message": error.message, "code": error.code}
                for error in output_data.errors
            ]

        formatted = {
            "Runtime statistics": runtime_statistics,
            "Errors": errors,
        }
        if output_data.skip_reason is not None:
            formatted["Skipped"] = output_data.skip_reason
        return formatted


class WorkRequestPlugin(abc.ABC):
    """
    WorkRequests with specific outputs must subclass it.

    When subclassing, the subclass:
    - Is automatically used by the /work-request/ID/ endpoints
    - Must define "template_name", "task_type", and "task_name"
    - Must implement "get_context_data()"
    """

    model = WorkRequest
    task_type: str
    task_name: str
    task: DBTask[Any, Any]
    template_name: str

    #: Maps (task_type, task_name) tuples to their associated plugin class.
    #: Used to look up the correct plugin for a given work request.
    _work_request_plugins: dict[tuple[str, str], type["WorkRequestPlugin"]] = {}

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:  # noqa: U100
        """Register the plugin."""
        cls._work_request_plugins[(cls.task_type, cls.task_name)] = cls

    def __init__(self, task: DBTask[Any]) -> None:
        """Store the work_request to display."""
        self.task = task
        self.work_request = task.work_request

    @classmethod
    def plugin_for(cls, task: DBTask[Any]) -> "WorkRequestPlugin | None":
        """Return WorkRequestPlugin for task_name or None."""
        plugin_class = cls._work_request_plugins.get(
            (task.TASK_TYPE, task.name)
        )
        if plugin_class is None:
            return None
        plugin = plugin_class(task)
        if not plugin.is_enabled():
            return None
        return plugin

    def is_enabled(self) -> bool:
        """Check if the tab should appear."""
        return True

    def add_tabs(self, tabs: list[Tab]) -> None:
        """Add extra tabs to the tab list."""

    def get_description_data(self) -> dict[str, Any]:
        """Return data for the description or {} if dynamic_data is None."""
        if self.task.dynamic_data is None:
            return {}

        return self.do_get_description_data()

    def do_get_description_data(self) -> dict[str, Any]:
        """Return data used for the description."""
        return {}

    def get_context_data(self) -> dict[str, Any]:
        """Must be implemented by subclasses."""
        raise NotImplementedError()


class WorkRequestListView(
    TableMixin[WorkRequest], WorkRequestObjectMixin, ListViewBase[WorkRequest]
):
    """List work requests."""

    model = WorkRequest
    table_class = WorkRequestTable
    template_name = "web/work_request-list.html"
    context_object_name = "work_request_list"
    title = "List of work requests"
    paginate_by = 50

    @override
    def get_place(self) -> Place:
        place = super().get_place()
        place.parent = context.require_workspace().ui(self.request).place
        return place

    @staticmethod
    def _filter_by_architecture(
        work_requests: WorkRequestQuerySet[WorkRequest], architecture: str
    ) -> WorkRequestQuerySet[WorkRequest]:
        return work_requests.filter(
            scheduler_tags_required__contains=[
                wtags.BUILD_ARCH_PREFIX + architecture
            ]
        )

    @override
    def get_queryset(self) -> WorkRequestQuerySet[WorkRequest]:
        """Filter work requests displayed by the workspace GET parameter."""
        queryset = super().get_queryset().exclude(task_type=TaskTypes.INTERNAL)

        arch_param = self.request.GET.get("arch")
        status_param = self.request.GET.get("status")

        status_mapping = {
            "pending": WorkRequest.Statuses.PENDING,
            "running": WorkRequest.Statuses.RUNNING,
            "completed": WorkRequest.Statuses.COMPLETED,
            "aborted": WorkRequest.Statuses.ABORTED,
            "blocked": WorkRequest.Statuses.BLOCKED,
        }

        status_value = (
            None if status_param is None else status_mapping.get(status_param)
        )

        if status_param is not None and status_value is None:
            messages.warning(
                self.request, 'Invalid "status" parameter, ignoring it'
            )

        if status_value:
            queryset = queryset.filter(status=status_value)

        if arch_param:
            if status_value == WorkRequest.Statuses.PENDING:
                # Filtering by architecture is only allowed for PENDING
                # Work Requests. Filtering by architecture, currently,
                # requires instantiating the Task and calling
                # build_architecture()
                queryset = self._filter_by_architecture(queryset, arch_param)
            else:
                messages.warning(
                    self.request,
                    'Filter by architecture is only supported when '
                    'also filtering by "status=pending", ignoring architecture'
                    'filtering',
                )

        return queryset


class WorkRequestRetryView(
    WorkRequestObjectMixin, SingleObjectMixinBase[WorkRequest], View
):
    """Form view for retrying a work request."""

    model = WorkRequest

    def retry(self, work_request: WorkRequest) -> HttpResponse:
        """Retry the work request."""
        try:
            new_work_request = work_request.retry()
        except CannotRetry as e:
            messages.error(self.request, f"Cannot retry: {e}")
            return redirect(work_request.get_absolute_url())

        return redirect(new_work_request.get_absolute_url())

    def post(
        self,
        request: HttpRequest,  # noqa: ARG002, U100
        *args: Any,
        **kwargs: Any,
    ) -> HttpResponse:
        """Handle POST requests."""
        work_request = self.get_object()
        self.enforce(work_request.can_retry)
        return self.retry(work_request)


class WorkRequestAbortView(
    WorkRequestObjectMixin, SingleObjectMixinBase[WorkRequest], View
):
    """Form view for aborting a work request."""

    model = WorkRequest

    def abort(self, work_request: WorkRequest) -> HttpResponse:
        """Abort the work request."""
        try:
            work_request.abort()
        except CannotAbort as e:
            messages.error(self.request, f"Cannot abort: {e}")

        return redirect(work_request.get_absolute_url())

    def post(self, *args: Any, **kwargs: Any) -> HttpResponse:
        """Handle POST requests."""
        work_request = self.get_object()
        self.enforce(work_request.can_abort)
        WorkRequest.objects.filter(
            id=work_request.id
        ).lock_workflows_for_update()
        return self.abort(work_request)


class WorkRequestUnblockView(
    WorkRequestObjectMixin,
    SingleObjectMixinBase[WorkRequest],
    FormMixinBase[WorkRequestUnblockForm],
    View,
):
    """Form view for reviewing a work request awaiting manual approval."""

    model = WorkRequest
    form_class = WorkRequestUnblockForm

    def get_action(self) -> WorkRequestManualUnblockAction | None:
        """Get the unblock action from query arguments."""
        raw_action = self.request.POST["action"]
        match raw_action:
            case "Accept":
                return WorkRequestManualUnblockAction.ACCEPT
            case "Reject":
                return WorkRequestManualUnblockAction.REJECT
            case "Record notes only":
                return None
            case _:
                raise HttpError400(f"Invalid action parameter: {raw_action!r}")

    def unblock(self, work_request: WorkRequest) -> HttpResponse:
        """Perform the unblock."""
        # This view requires a permission, so the user can't be an
        # AnonymousUser by this point.
        assert isinstance(self.request.user, User)

        form = self.get_form()
        if not form.is_valid():
            # TODO: This is ugly and we should probably try to render
            # work_requests:detail with inline unblock form errors instead.
            # However, since the form is so limited it's very difficult to
            # actually hit this case in practice, so it doesn't seem worth
            # the effort for now.
            return render(
                self.request,
                "400.html",
                context={"error": form.errors.as_json()},
                status=status.HTTP_400_BAD_REQUEST,
            )

        try:
            work_request.review_manual_unblock(
                user=self.request.user,
                notes=form.cleaned_data["notes"],
                action=self.get_action(),
            )
        except CannotUnblock as e:
            raise HttpError400(f"Cannot unblock: {e}")
        return redirect(work_request.get_absolute_url())

    @catch_http_errors
    def post(
        self,
        request: HttpRequest,  # noqa: ARG002, U100
        *args: Any,
        **kwargs: Any,
    ) -> HttpResponse:
        """Handle POST requests."""
        work_request = self.get_object()
        self.enforce(work_request.can_unblock)
        return self.unblock(work_request)


class WorkRequestConfirmView(
    WorkRequestObjectMixin,
    SingleObjectMixinBase[WorkRequest],
    FormMixinBase[WorkRequestConfirmForm],
    View,
):
    """Form view for confirming a work request."""

    model = WorkRequest
    form_class = WorkRequestConfirmForm

    @property
    def should_confirm(self) -> bool:
        """Get whether to confirm the work request from query arguments."""
        action = self.request.POST["action"]
        match action:
            case "confirm":
                return True
            case "deny":
                return False
            case _:
                raise HttpError400(f"Invalid action parameter: {action!r}")

    def confirm_or_deny(self, work_request: WorkRequest) -> HttpResponse:
        """Perform the confirmation or denial."""
        # This view requires a permission, so the user can't be an
        # AnonymousUser by this point.
        assert isinstance(self.request.user, User)

        if (
            work_request.task_type != TaskTypes.WAIT
            or work_request.task_name != "confirm"
        ):
            raise HttpError400(
                f"Expected work request to be Wait/Confirm; got "
                f"{work_request.task_type}/{work_request.task_name}"
            )
        task = work_request.get_task()
        assert isinstance(task, Confirm)

        form = self.get_form()
        if not form.is_valid():
            # TODO: This is ugly and we should probably try to render
            # work_requests:detail with inline confirm form errors instead.
            # However, since the form is so limited it's very difficult to
            # actually hit this case in practice, so it doesn't seem worth
            # the effort for now.
            return render(
                self.request,
                "400.html",
                context={"error": form.errors.as_json()},
                status=status.HTTP_400_BAD_REQUEST,
            )

        task.respond(
            user=self.request.user,
            confirm=self.should_confirm,
            comment=form.cleaned_data["comment"],
        )
        return redirect(work_request.get_absolute_url())

    @catch_http_errors
    def post(
        self,
        request: HttpRequest,  # noqa: ARG002, U100
        *args: Any,
        **kwargs: Any,
    ) -> HttpResponse:
        """Handle POST requests."""
        work_request = self.get_object()
        self.enforce(work_request.can_confirm)
        return self.confirm_or_deny(work_request)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Default Django settings for the Debusine signing project.

Most settings are documented in this file and they are initialized to some
reasonable default values when possible.  They will be extended (and
possibly overridden) by settings from the other modules in this package
depending on the setup selected by the administrator. You likely won't
have to modify that file.

You should instead create local.py to put your site-specific settings (use
local.py.sample as template).

Here are the most important settings:

:py:data:`DEBUSINE_DATA_PATH`
    The directory where debusine will hold its data. The directory is
    further sub-divided in multiple directories for specific use
    cases (e.g. cache, keyring, static, media, templates, etc.).
    Defaults to the "data" sub-directory in the debusine
    base directory (where the code lives).

Some settings have default values which are computed dynamically from
other settings. Those settings can also be overridden. Here's the list
of those settings.

More settings:

"""

import os.path
from collections.abc import Callable
from os.path import dirname
from pathlib import Path
from typing import Any

import django
from django.core.exceptions import ImproperlyConfigured

from debusine.signing.signing_utils import read_private_key

if django.VERSION < (4, 2):
    raise ImproperlyConfigured("Debusine needs Django >= 4.2")

# Django's debug mode, never enable this in production
DEBUG = False

BASE_DIR = dirname(dirname(dirname(dirname(__file__))))
DEBUSINE_DATA_PATH = os.path.join(BASE_DIR, 'data')

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# In a Windows environment this must be set to your system time zone.
TIME_ZONE = 'Etc/UTC'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = True

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

INSTALLED_APPS: list[str] = [
    'debusine.signing',
    'debusine.signing.db',
]

# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s [%(module)s/%(process)d/%(thread)d] '
            + '%(levelname)s: %(message)s'
        },
        'standard': {
            'format': '%(asctime)s %(process)d %(levelname)s: %(message)s'
        },
        'simple': {'format': '%(asctime)s %(levelname)s: %(message)s'},
        'messageonly': {
            "format": "%(message)s",
        },
    },
    'filters': {
        'require_debug_false': {'()': 'django.utils.log.RequireDebugFalse'},
        'require_debug_true': {'()': 'django.utils.log.RequireDebugTrue'},
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
            'filters': ['require_debug_true'],
        },
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler',
        },
        'null': {'level': 'DEBUG', 'class': 'logging.NullHandler'},
        'journal': {
            'level': 'DEBUG',
            'class': 'systemd.journal.JournalHandler',
            'formatter': 'messageonly',
        },
    },
    'loggers': {
        'root': {
            'handlers': ['journal'],
            'level': 'DEBUG',
            'propagate': False,
        },
        'django.request': {
            'handlers': ['journal', 'mail_admins'],
            'level': 'ERROR',
            'propagate': False,
        },
        'django.security': {
            'handlers': ['journal', 'mail_admins'],
            'level': 'ERROR',
            'propagate': False,
        },
        'django.security.DisallowedHost': {
            'handlers': ['journal'],
            'level': 'ERROR',
            'propagate': False,
        },
        'py.warnings': {'handlers': ['console']},
    },
}

# === Debusine specific settings ===

# Encryption keys used by the signing service to encrypt private keys; the
# first in the list is used for newly-encrypted private keys, while the
# others may also be used for decryption
DEBUSINE_SIGNING_PRIVATE_KEYS = [
    read_private_key(path)
    for path in sorted(Path(__file__).parent.glob("*.key"))
]

# Allow signing objects that declare a trust chain to these certificate
# fingerprints.  See
# https://wiki.debian.org/SecureBoot/Discussion#Describing_the_trust_chain.
DEBUSINE_SIGNING_TRUSTED_CERTS: list[str] = []

# The lambda functions are evaluated at the end of the settings import
# logic. They provide default values to settings which have not yet been
# set (neither above nor in local.py).
_COMPUTE_DEFAULT_SETTINGS: tuple[
    tuple[str, Callable[[dict[str, Any]], str]], ...
] = ()


def compute_default_settings(target: dict[str, Any]) -> None:
    """
    Dynamically generate some default settings.

    There are many settings whose default value depends on another
    setting. They are defined in the _COMPUTE_DEFAULT_SETTINGS dict
    with a function that evaluates them.
    """
    for setting, value in _COMPUTE_DEFAULT_SETTINGS:
        if setting in target:
            continue  # Settings is already defined
        target[setting] = value(target)
    # Update DATABASES with full paths
    dbconf = target['DATABASES']['default']
    if dbconf['ENGINE'] == 'django.db.backends.sqlite3':
        if '/' not in dbconf['NAME']:
            dbconf['NAME'] = os.path.join(
                target['DEBUSINE_DATA_PATH'], dbconf['NAME']
            )
        if (
            'TEST' in dbconf
            and 'NAME' in dbconf['TEST']
            and '/' not in dbconf['TEST']['NAME']
        ):
            dbconf['TEST']['NAME'] = os.path.join(
                target['DEBUSINE_DATA_PATH'], dbconf['TEST']['NAME']
            )


LOGGING_CONFIG = "debusine.django.django_utils.test_data_override"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for utils module."""

import hashlib
import stat
from importlib.metadata import PackageNotFoundError
from pathlib import Path
from unittest import mock

from debian import deb822
from requests.structures import CaseInsensitiveDict

from debusine.test import TestCase
from debusine.utils import (
    CALCULATE_HASH_CHUNK_SIZE,
    DjangoChoicesEnum,
    atomic_writer,
    calculate_hash,
    debusine_version,
    find_file_suffixes,
    find_files_suffixes,
    get_user_agent,
    is_command_available,
    parse_content_range_header,
    parse_range_header,
    read_changes,
    read_dsc,
)


class TestUtilsTests(TestCase):
    """Tests for the test_utils functions."""

    def test_calculate_hash_small_file(self) -> None:
        """calculate_hash return the correct sha256 for the file."""
        contents = b"debusine"
        sha256_expected = hashlib.sha256(contents).digest()

        temporary_file = self.create_temporary_file(contents=contents)

        self.assertEqual(
            calculate_hash(Path(temporary_file), "sha256"),
            sha256_expected,
        )

    def test_calculate_hash_big_file(self) -> None:
        """
        calculate_hash return the correct sha256 for a big file.

        A big file is a file bigger than CALCULATE_HASH_CHUNK_SIZE.
        """
        size_to_write = int(CALCULATE_HASH_CHUNK_SIZE * 2.5)

        contents = bytearray(size_to_write)

        temporary_file = self.create_temporary_file(contents=contents)

        self.assertEqual(
            calculate_hash(Path(temporary_file), "sha256"),
            hashlib.sha256(contents).digest(),
        )

    def test_parse_range_header_no_header(self) -> None:
        """parse_range_header return None: no "Range" header."""
        self.assertIsNone(parse_range_header(CaseInsensitiveDict({}), 30))

    def test_parse_range_header_invalid_header(self) -> None:
        """parse_range_header raise ValueError: invalid header."""
        invalid_header = "invalid-header"
        error = f'Invalid Range header: "{invalid_header}"'
        with self.assertRaisesRegex(ValueError, error):
            parse_range_header(
                CaseInsensitiveDict({"Range": invalid_header}), 30
            )

    def test_parse_range_header_no_positions(self) -> None:
        """parse_range_header raise ValueError: neither start nor end."""
        invalid_header = "bytes=-"
        error = f'Invalid Range header: "{invalid_header}"'
        with self.assertRaisesRegex(ValueError, error):
            parse_range_header(
                CaseInsensitiveDict({"Range": invalid_header}), 30
            )

    def test_parse_range_header(self) -> None:
        """parse_range_header return dictionary with start and end."""
        for range_header, expected in (
            ("bytes=10-24", {"start": 10, "end": 24}),
            ("bytes=10-", {"start": 10, "end": 29}),
            ("bytes=-10", {"start": 20, "end": 29}),
        ):
            with self.subTest(range_header=range_header):
                headers = CaseInsensitiveDict({"Range": range_header})
                self.assertEqual(parse_range_header(headers, 30), expected)

    def test_parse_content_range_header_no_header(self) -> None:
        """parse_content_range_header return None: no "Content-Range"."""
        self.assertIsNone(parse_content_range_header({}))

    def test_parse_content_range_header_start_end_size(self) -> None:
        """parse_content_range_header return dict with integers."""
        for header, expected in (
            ("bytes 10-20/30", {"start": 10, "end": 20, "size": 30}),
            ("bytes 10-20/*", {"start": 10, "end": 20, "size": "*"}),
            ("bytes */30", {"start": "*", "end": None, "size": 30}),
            ("bytes */*", {"start": "*", "end": None, "size": "*"}),
        ):
            with self.subTest(header=header):
                headers = {"Content-Range": header}

                self.assertEqual(parse_content_range_header(headers), expected)

    def test_parse_content_range_header_invalid_header(self) -> None:
        """parse_content_range_header raise ValueError: invalid header."""
        invalid_header = "invalid-header"
        error = f'Invalid Content-Range header: "{invalid_header}"'
        with self.assertRaisesRegex(ValueError, error):
            parse_content_range_header({"Content-Range": invalid_header})

    def test_read_dsc_file(self) -> None:
        """_read_dsc return a deb822.Dsc object: it was a valid Dsc file."""
        dsc_file = self.create_temporary_file(suffix=".dsc")
        self.write_dsc_example_file(dsc_file)

        dsc = read_dsc(dsc_file)

        self.assertIsInstance(dsc, deb822.Dsc)

    def test_read_dsc_file_no_source_field(self) -> None:
        dsc_file = self.create_temporary_file(contents=b"invalid dsc file")

        with self.assertRaisesRegex(
            ValueError, f"{dsc_file.name!r} does not contain Source field"
        ):
            read_dsc(dsc_file)

    def test_read_dsc_file_no_version_field(self) -> None:
        dsc_file = self.create_temporary_file(contents=b"Source: hello\n")

        with self.assertRaisesRegex(
            ValueError, f"{dsc_file.name!r} does not contain Version field"
        ):
            read_dsc(dsc_file)

    def test_read_dsc_file_multiple_path_components(self) -> None:
        dsc_file = self.create_temporary_file(suffix=".dsc")

        for field, data in (
            ("Files", {"Files": "\n 0000 1 ../file"}),
            ("Checksums-Sha1", {"Checksums-Sha1": "\n 0000 1 ../file"}),
            ("Checksums-Sha256", {"Checksums-Sha256": "\n 0000 1 ../file"}),
        ):
            with self.subTest(field=field):
                with dsc_file.open("w") as f:
                    deb822.Dsc(
                        {"Source": "hello", "Version": "1.0", **data}
                    ).dump(f, text_mode=True)

                with self.assertRaisesRegex(
                    ValueError,
                    rf"{dsc_file.name!r} has an entry in {field} with more"
                    rf" than one path component: '\.\./file'",
                ):
                    read_dsc(dsc_file)

    def test_read_changes_file(self) -> None:
        """read_changes returns deb822.Changes for a valid .changes file."""
        changes_file = self.create_temporary_file(suffix=".changes")
        self.write_changes_file(changes_file, [])

        changes = read_changes(changes_file)

        self.assertIsInstance(changes, deb822.Changes)

    def test_read_changes_file_multiple_path_components(self) -> None:
        changes_file = self.create_temporary_file(suffix=".changes")

        for field, data in (
            ("Files", {"Files": "\n 0000 1 devel optional ../file"}),
            ("Checksums-Sha1", {"Checksums-Sha1": "\n 0000 1 ../file"}),
            ("Checksums-Sha256", {"Checksums-Sha256": "\n 0000 1 ../file"}),
        ):
            with self.subTest(field=field):
                with changes_file.open("w") as f:
                    deb822.Dsc(
                        {"Source": "hello", "Version": "1.0", **data}
                    ).dump(f, text_mode=True)

                with self.assertRaisesRegex(
                    ValueError,
                    rf"{changes_file.name!r} has an entry in {field} with more"
                    rf" than one path component: '\.\./file'",
                ):
                    read_changes(changes_file)

    def test_find_file_suffixes_multiple_files(self) -> None:
        """find_file_suffixes() return two files with different suffixes."""
        temporary_directory = self.create_temporary_directory()

        # Add a deb file
        (deb := temporary_directory / "file.deb").write_text("deb")

        # Add a udeb file
        (udeb := temporary_directory / "file.udeb").write_text("deb")

        # Add a file that is not returned
        (temporary_directory / "README.txt").write_text("README")

        # Check deb and udeb files
        self.assertEqual(
            find_files_suffixes(temporary_directory, [".deb", ".udeb"]),
            sorted([deb, udeb]),
        )

    def test_find_file_suffix_raise_runtime_error(self) -> None:
        """find_file() raise RuntimeError: two possible log files."""
        temporary_directory = self.create_temporary_directory()

        suffix = "build"

        (file1 := temporary_directory / f"log1.{suffix}").write_text("Log1")
        (file2 := temporary_directory / f"log2.{suffix}").write_text("Log2")

        # assertRaisesRegex must have the chars [] escaped
        files = (
            str(sorted([str(file1), str(file2)]))
            .replace("[", "")
            .replace("]", "")
        )

        with self.assertRaisesRegex(
            RuntimeError, rf"^More than one \['{suffix}'\] file: \[{files}\]$"
        ):
            find_file_suffixes(temporary_directory, [suffix])

    def test_find_files_suffix(self) -> None:
        """find_files_suffix() return multiple files."""
        temporary_directory = self.create_temporary_directory()

        # Create two files to be returned
        (deb1_file := temporary_directory / "hello.deb").write_text("First")
        (deb2_file := temporary_directory / "hello-2.deb").write_text("Second")

        # Create a non-relevant file
        (temporary_directory / "README.txt").write_text("a test")

        # Create a file that could match (*.deb) but is a symbolic link
        (temporary_directory / "temp.deb").symlink_to(deb1_file)

        # Create a file that could match (*.deb) but is a directory
        (temporary_directory / "debian.deb").mkdir()

        self.assertEqual(
            find_files_suffixes(temporary_directory, [".deb"]),
            sorted([deb1_file, deb2_file]),
        )

    def test_find_files_suffix_include_symbolic_links(self) -> None:
        """
        find_files_suffix() return files including symbolic links.

        The option include_symlinks is True.
        """
        temporary_directory = self.create_temporary_directory()

        # Create one file and one symbolic link to the file
        file_name = "file.txt"
        symbolic_link_name = "symbolic.txt"
        (file := temporary_directory / file_name).write_text("test")
        (symbolic_link := temporary_directory / symbolic_link_name).symlink_to(
            file
        )

        self.assertEqual(
            find_files_suffixes(
                temporary_directory, [".txt"], include_symlinks=True
            ),
            [file, symbolic_link],
        )

    def test_find_file_suffixes_one_file(self) -> None:
        """find_file_suffixes() return the expected file."""
        temporary_directory = self.create_temporary_directory()

        # Add a log file
        (log_file := temporary_directory / "log.build").write_text("log")

        # Add a file that is not returned
        (temporary_directory / "README.txt").write_text("README")

        # Check return the correct log file
        self.assertEqual(
            find_file_suffixes(temporary_directory, [".build"]),
            log_file,
        )

    def patch_shutil_which(self, available: bool) -> None:
        """Patch shutil.which to respond either positively or negatively."""
        patcher = mock.patch(
            "shutil.which", return_value="/usr/bin/test" if available else None
        )
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_is_command_available_yes(self) -> None:
        """is_command_available returns True if the command exists."""
        self.patch_shutil_which(True)
        self.assertTrue(is_command_available("test"))

    def test_is_command_available_no(self) -> None:
        """is_command_available returns False if the command does not exist."""
        self.patch_shutil_which(False)
        self.assertFalse(is_command_available("test"))

    def test_choices(self) -> None:
        """Test that .choices returns the values in Django choices pairs."""

        class TestChoices(DjangoChoicesEnum):
            FOO = "foo"
            BAR = "bar"

        self.assertEqual(
            list(TestChoices.choices), [("foo", "foo"), ("bar", "bar")]
        )

    def test_atomic_writer(self) -> None:
        """``atomic_writer`` writes to the target path."""
        temp_dir = self.create_temporary_directory()
        path = temp_dir / "file"

        with atomic_writer(path) as fd:
            fd.write(b"contents\n")
            self.assertFalse(path.exists())
            [temp_path] = temp_dir.iterdir()
            self.assertEqual(temp_path.parent, temp_dir)
            self.assertRegex(temp_path.name, r"^file.*\.new$")

        self.assertTrue(path.exists())
        self.assertEqual(path.read_bytes(), b"contents\n")
        self.assertEqual(stat.S_IMODE(path.stat().st_mode), 0o600)
        self.assertEqual(list(temp_dir.iterdir()), [path])

    def test_atomic_writer_text(self) -> None:
        """``atomic_writer`` can write to the target path as text."""
        temp_dir = self.create_temporary_directory()
        path = temp_dir / "file"

        with atomic_writer(path, mode="w") as fd:
            fd.write("contents\n")
            self.assertFalse(path.exists())
            [temp_path] = temp_dir.iterdir()
            self.assertEqual(temp_path.parent, temp_dir)
            self.assertRegex(temp_path.name, r"^file.*\.new$")

        self.assertTrue(path.exists())
        self.assertEqual(path.read_text(), "contents\n")
        self.assertEqual(stat.S_IMODE(path.stat().st_mode), 0o600)
        self.assertEqual(list(temp_dir.iterdir()), [path])

    def test_atomic_writer_chmod(self) -> None:
        """``atomic_writer`` can set the target path's permissions."""
        temp_dir = self.create_temporary_directory()
        path = temp_dir / "file"

        with atomic_writer(path, chmod=0o644) as fd:
            fd.write(b"contents\n")
            self.assertFalse(path.exists())
            [temp_path] = temp_dir.iterdir()
            self.assertEqual(temp_path.parent, temp_dir)
            self.assertRegex(temp_path.name, r"^file.*\.new$")

        self.assertTrue(path.exists())
        self.assertEqual(path.read_bytes(), b"contents\n")
        self.assertEqual(stat.S_IMODE(path.stat().st_mode), 0o644)
        self.assertEqual(list(temp_dir.iterdir()), [path])

    def test_atomic_writer_exceptions(self) -> None:
        """``atomic_writer`` cleans up after exceptions."""
        temp_dir = self.create_temporary_directory()
        path = temp_dir / "file"

        with self.assertRaisesRegex(RuntimeError, "Boom"), atomic_writer(path):
            raise RuntimeError("Boom")

        self.assertEqual(list(temp_dir.iterdir()), [])

    def test_debusine_version(self) -> None:
        """``debusine_version`` returns debusine's version if known."""
        with mock.patch(
            "debusine.utils.version", return_value="0.1.0"
        ) as mock_version:
            self.assertEqual(debusine_version(), "0.1.0")

        mock_version.assert_called_once_with("debusine")

    def test_debusine_version_not_installed(self) -> None:
        """``debusine_version`` tolerates not knowing debusine's version."""
        with mock.patch(
            "debusine.utils.version", side_effect=PackageNotFoundError
        ):
            self.assertIsNone(debusine_version())

    def test_get_user_agent_with_version(self) -> None:
        """``get_user_agent`` returns debusine/version when known."""
        with mock.patch("debusine.utils.debusine_version", return_value="1.2"):
            self.assertEqual(get_user_agent(), "debusine/1.2")

    def test_get_user_agent_without_version(self) -> None:
        """``get_user_agent`` returns debusine when version is unknown."""
        with mock.patch("debusine.utils.debusine_version", return_value=None):
            self.assertEqual(get_user_agent(), "debusine")
